# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


"""Builds the bus admittance matrix and branch admittance matrices.
"""

from numpy import ones, conj, nonzero, any, exp, pi, hstack, real, int64, errstate
from scipy.sparse import csr_matrix

from pandapower.pypower.idx_brch import F_BUS, T_BUS, BR_R, BR_X, BR_B, BR_G, BR_STATUS, SHIFT, TAP, BR_R_ASYM, \
    BR_X_ASYM, BR_G_ASYM, BR_B_ASYM
from pandapower.pypower.idx_bus import GS, BS


def makeYbus(baseMVA, bus, branch):
    """Builds the bus admittance matrix and branch admittance matrices.

    Returns the full bus admittance matrix (i.e. for all buses) and the
    matrices C{Yf} and C{Yt} which, when multiplied by a complex voltage
    vector, yield the vector currents injected into each line from the
    "from" and "to" buses respectively of each line. Does appropriate
    conversions to p.u.

    @see: L{makeSbus}

    @author: Ray Zimmerman (PSERC Cornell)
    @author: Richard Lincoln
    """
    ## constants
    nb = bus.shape[0]  ## number of buses
    nl = branch.shape[0]  ## number of lines

    ## for each branch, compute the elements of the branch admittance matrix where
    ##
    ##      | If |   | Yff  Yft |   | Vf |
    ##      |    | = |          | * |    |
    ##      | It |   | Ytf  Ytt |   | Vt |
    ##
    Ytt, Yff, Yft, Ytf = branch_vectors(branch, nl)
    ## compute shunt admittance
    ## if Psh is the real power consumed by the shunt at V = 1.0 p.u.
    ## and Qsh is the reactive power injected by the shunt at V = 1.0 p.u.
    ## then Psh - j Qsh = V * conj(Ysh * V) = conj(Ysh) = Gs - j Bs,
    ## i.e. Ysh = Psh + j Qsh, so ...
    ## vector of shunt admittances
    Ysh = (bus[:, GS] + 1j * bus[:, BS]) / baseMVA

    ## build connection matrices
    f = real(branch[:, F_BUS]).astype(int64)  ## list of "from" buses
    t = real(branch[:, T_BUS]).astype(int64)  ## list of "to" buses
    ## connection matrix for line & from buses
    Cf = csr_matrix((ones(nl), (range(nl), f)), (nl, nb))
    ## connection matrix for line & to buses
    Ct = csr_matrix((ones(nl), (range(nl), t)), (nl, nb))

    ## build Yf and Yt such that Yf * V is the vector of complex branch currents injected
    ## at each branch's "from" bus, and Yt is the same for the "to" bus end
    i = hstack([range(nl), range(nl)])  ## double set of row indices

    Yf = csr_matrix((hstack([Yff, Yft]), (i, hstack([f, t]))), (nl, nb))
    Yt = csr_matrix((hstack([Ytf, Ytt]), (i, hstack([f, t]))), (nl, nb))
    # Yf = spdiags(Yff, 0, nl, nl) * Cf + spdiags(Yft, 0, nl, nl) * Ct
    # Yt = spdiags(Ytf, 0, nl, nl) * Cf + spdiags(Ytt, 0, nl, nl) * Ct

    ## build Ybus
    Ybus = Cf.T * Yf + Ct.T * Yt + \
           csr_matrix((Ysh, (range(nb), range(nb))), (nb, nb))

    # for canonical format
    for Y in (Ybus, Yf, Yt):
        Y.eliminate_zeros()
        Y.sum_duplicates()
        Y.sort_indices()
        del Y._has_canonical_format

    return Ybus, Yf, Yt


@errstate(all="raise")
def branch_vectors(branch, nl):
    stat = branch[:, BR_STATUS]  # ones at in-service branches
    Ysf = stat / (branch[:, BR_R] + 1j * branch[:, BR_X])  # series admittance
    if any(branch[:, BR_R_ASYM]) or any(branch[:, BR_X_ASYM]):
        Yst = stat / (branch[:, BR_R] + branch[:, BR_R_ASYM] +
                      1j * (branch[:, BR_X] + branch[:, BR_X_ASYM]))
    else:
        Yst = Ysf

    Bcf = stat * (branch[:, BR_G] + 1j * branch[:, BR_B])  # branch charging admittance
    if any(branch[:, BR_G_ASYM]) or any(branch[:, BR_B_ASYM]):
        Bct = stat * (branch[:, BR_G] + branch[:, BR_G_ASYM] +
                      1j * (branch[:, BR_B] + branch[:, BR_B_ASYM]))
    else:
        Bct = Bcf

    tap = ones(nl)  # default tap ratio = 1
    i = nonzero(real(branch[:, TAP]))  # indices of non-zero tap ratios
    tap[i] = real(branch[i, TAP])  # assign non-zero tap ratios
    tap = tap * exp(1j * pi / 180 * branch[:, SHIFT])  # add phase shifters

    Ytt = Yst + Bct / 2
    Yff = (Ysf + Bcf / 2) / (tap * conj(tap))
    Yft = - Ysf / conj(tap)
    Ytf = - Yst / tap
    return Ytt, Yff, Yft, Ytf




# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np
import pandas as pd

from pandapower.pypower.idx_brch_sc import IKSS_F, IKSS_T, IP_F, IP_T, ITH_F, ITH_T, PKSS_F, QKSS_F, PKSS_T, QKSS_T, \
    VKSS_MAGN_F, VKSS_MAGN_T, VKSS_ANGLE_F, VKSS_ANGLE_T, IKSS_ANGLE_F, IKSS_ANGLE_T
from pandapower.pypower.idx_bus_sc import IKSS1, IP, ITH, IKSS2, R_EQUIV_OHM, X_EQUIV_OHM, SKSS
from pandapower.pypower.idx_bus import BUS_TYPE, BASE_KV
from pandapower.results_branch import _copy_switch_results_from_branches
from pandapower.results import BRANCH_RESULTS_KEYS


def _copy_result_to_ppci_orig(ppci_orig, ppci, ppci_bus, calc_options):
    if ppci_orig is ppci:
        return

    ppci_orig["bus"][ppci_bus, :] = ppci["bus"][ppci_bus, :]
    if calc_options["branch_results"]:
        if calc_options["return_all_currents"]:
            ppci_orig["internal"]["br_res_ks_ppci_bus"] =\
                ppci_bus if "br_res_ks_ppci_bus" not in ppci_orig["internal"]\
                else np.r_[ppci_orig["internal"]["br_res_ks_ppci_bus"], ppci_bus]

            for res_key in BRANCH_RESULTS_KEYS:
                # Skip not required data points
                if res_key not in ppci["internal"]:
                    continue

                if res_key not in ppci_orig["internal"]:
                    ppci_orig["internal"][res_key] = ppci["internal"][res_key]
                else:
                    ppci_orig["internal"][res_key] = np.c_[ppci_orig["internal"][res_key],
                                                           ppci["internal"][res_key]]
        else:
            case = calc_options["case"]
            branch_results_cols = [IKSS_F, IKSS_T, IP_F, IP_T, ITH_F, ITH_T]
            # added new calculation values:
            branch_results_cols_add = [IKSS_ANGLE_F, IKSS_ANGLE_T, PKSS_F, QKSS_F, PKSS_T, QKSS_T,
                                       VKSS_MAGN_F, VKSS_MAGN_T, VKSS_ANGLE_F, VKSS_ANGLE_T]
            if case == "max":
                ppci_orig["branch"][:, branch_results_cols] =\
                    np.maximum(np.nan_to_num(ppci["branch"][:, branch_results_cols]),
                               np.nan_to_num(ppci_orig["branch"][:, branch_results_cols]))
                # excluding new values from nan to num
                ppci_orig["branch"][:, branch_results_cols_add] = ppci["branch"][:, branch_results_cols_add]
            else:
                ppci_orig["branch"][:, branch_results_cols] =\
                    np.minimum(np.nan_to_num(ppci["branch"][:, branch_results_cols], nan=1e10),
                               np.nan_to_num(ppci_orig["branch"][:, branch_results_cols], nan=1e10))
                # excluding new values from nan to num
                ppci_orig["branch"][:, branch_results_cols_add] = ppci["branch"][:, branch_results_cols_add]


def _get_bus_ppc_idx_for_br_all_results(net, ppc, bus):
    bus_lookup = net._pd2ppc_lookups["bus"]
    if bus is None:
        bus = net.bus.index

    ppc_index = bus_lookup[bus]
    ppc_index[ppc["bus"][ppc_index, BUS_TYPE] == 4] = -1
    return bus, ppc_index


def _extract_results(net, ppc, ppc_0, bus):
    _get_bus_results(net, ppc, ppc_0, bus)
    if net._options["branch_results"]:
        if net._options['return_all_currents']:
            _get_line_all_results(net, ppc, bus)
            _get_trafo_all_results(net, ppc, bus)
            _get_trafo3w_all_results(net, ppc, bus)
            _get_switch_all_results(net, ppc, bus)
        else:
            _get_line_results(net, ppc)
            _get_trafo_results(net, ppc)
            _get_trafo3w_results(net, ppc)
            _get_switch_results(net, ppc)


def _get_bus_results(net, ppc, ppc_0, bus):
    bus_lookup = net._pd2ppc_lookups["bus"]
    ppc_index = bus_lookup[net.bus.index]

    if net["_options"]["fault"] == "1ph":
        net.res_bus_sc["ikss_ka"] = ppc_0["bus"][ppc_index, IKSS1] + ppc["bus"][ppc_index, IKSS2]
        net.res_bus_sc["rk0_ohm"] = ppc_0["bus"][ppc_index, R_EQUIV_OHM]
        net.res_bus_sc["xk0_ohm"] = ppc_0["bus"][ppc_index, X_EQUIV_OHM]
        # in trafo3w, we add very high numbers (1e10) as impedances to block current
        # here, we need to replace such high values by np.inf
        baseZ = ppc_0["bus"][ppc_index, BASE_KV] ** 2 / ppc_0["baseMVA"]
        net.res_bus_sc.loc[net.res_bus_sc["xk0_ohm"]/baseZ > 1e9, "xk0_ohm"] = np.inf
        net.res_bus_sc.loc[net.res_bus_sc["rk0_ohm"]/baseZ > 1e9, "rk0_ohm"] = np.inf
    else:
        net.res_bus_sc["ikss_ka"] = ppc["bus"][ppc_index, IKSS1] + ppc["bus"][ppc_index, IKSS2]
        net.res_bus_sc["skss_mw"] = ppc["bus"][ppc_index, SKSS]
    if net._options["ip"]:
        net.res_bus_sc["ip_ka"] = ppc["bus"][ppc_index, IP]
    if net._options["ith"]:
        net.res_bus_sc["ith_ka"] = ppc["bus"][ppc_index, ITH]

    # Export also equivalent rk, xk on the calculated bus
    net.res_bus_sc["rk_ohm"] = ppc["bus"][ppc_index, R_EQUIV_OHM]
    net.res_bus_sc["xk_ohm"] = ppc["bus"][ppc_index, X_EQUIV_OHM]
    # if for some reason (e.g. contribution of ext_grid set close to 0) we used very high values for rk, xk, we replace them by np.inf
    baseZ = ppc["bus"][ppc_index, BASE_KV] ** 2 / ppc["baseMVA"]
    net.res_bus_sc.loc[net.res_bus_sc["rk_ohm"] / baseZ > 1e9, "rk_ohm"] = np.inf
    net.res_bus_sc.loc[net.res_bus_sc["xk_ohm"] / baseZ > 1e9, "xk_ohm"] = np.inf

    net.res_bus_sc = net.res_bus_sc.loc[bus, :]


def _get_line_results(net, ppc):
    branch_lookup = net._pd2ppc_lookups["branch"]
    case = net._options["case"]
    if "line" in branch_lookup:
        f, t = branch_lookup["line"]
        minmax = np.max if case == "max" else np.min
        net.res_line_sc["ikss_ka"] = minmax(ppc["branch"][f:t, [IKSS_F, IKSS_T]].real, axis=1)
        net.res_line_sc["ikss_from_ka"] = ppc["branch"][f:t, IKSS_F].real
        net.res_line_sc["ikss_from_degree"] = ppc["branch"][f:t, IKSS_ANGLE_F].real
        net.res_line_sc["ikss_to_ka"] = ppc["branch"][f:t, IKSS_T].real
        net.res_line_sc["ikss_to_degree"] = ppc["branch"][f:t, IKSS_ANGLE_T].real

        # adding columns for new calculated VPQ
        net.res_line_sc["p_from_mw"] = ppc["branch"][f:t, PKSS_F].real
        net.res_line_sc["q_from_mvar"] = ppc["branch"][f:t, QKSS_F].real

        net.res_line_sc["p_to_mw"] = ppc["branch"][f:t, PKSS_T].real
        net.res_line_sc["q_to_mvar"] = ppc["branch"][f:t, QKSS_T].real

        net.res_line_sc["vm_from_pu"] = ppc["branch"][f:t, VKSS_MAGN_F].real
        net.res_line_sc["va_from_degree"] = ppc["branch"][f:t, VKSS_ANGLE_F].real

        net.res_line_sc["vm_to_pu"] = ppc["branch"][f:t, VKSS_MAGN_T].real
        net.res_line_sc["va_to_degree"] = ppc["branch"][f:t, VKSS_ANGLE_T].real

        if net._options["ip"]:
            net.res_line_sc["ip_ka"] = minmax(ppc["branch"][f:t, [IP_F, IP_T]].real, axis=1)
        if net._options["ith"]:
            net.res_line_sc["ith_ka"] = minmax(ppc["branch"][f:t, [ITH_F, ITH_T]].real, axis=1)


def _get_switch_results(net, ppc):
    if len(net.switch) == 0:
        return
    branch_lookup = net._pd2ppc_lookups["branch"]
    case = net._options["case"]
    if "switch" in branch_lookup:
        f, t = branch_lookup["switch"]
        minmax = np.max if case == "max" else np.min
        bb_switches = net._impedance_bb_switches
        net.res_switch_sc.loc[bb_switches, "ikss_ka"] = minmax(ppc["branch"][f:t, [IKSS_F, IKSS_T]].real, axis=1)
        if net._options["ip"]:
            net.res_switch_sc.loc[bb_switches, "ip_ka"] = minmax(ppc["branch"][f:t, [IP_F, IP_T]].real, axis=1)
        if net._options["ith"]:
            net.res_switch_sc.loc[bb_switches, "ith_ka"] = minmax(ppc["branch"][f:t, [ITH_F, ITH_T]].real, axis=1)
    _copy_switch_results_from_branches(net, suffix="_sc", current_parameter="ikss_ka")
    if "in_ka" in net.switch.columns:
        net.res_switch_sc["loading_percent"] = net.res_switch_sc["ikss_ka"].values / net.switch["in_ka"].values * 100


def _get_branch_result_from_internal(variable, ppc, ppc_index, f, t):
    if variable in ppc["internal"]:
        return ppc["internal"][variable].iloc[f:t, :].loc[:, ppc_index].values.real.reshape(-1, 1)
    else:
        return np.full((t-f) * len(ppc_index), np.nan, dtype=np.float64)


def _get_line_all_results(net, ppc, bus):
    case = net._options["case"]

    bus, ppc_index = _get_bus_ppc_idx_for_br_all_results(net, ppc, bus)
    branch_lookup = net._pd2ppc_lookups["branch"]

    multindex = pd.MultiIndex.from_product([net.res_line_sc.index, bus], names=['line','bus'])
    net.res_line_sc = net.res_line_sc.reindex(multindex)

    if "line" in branch_lookup:
        f, t = branch_lookup["line"]
        minmax = np.maximum if case == "max" else np.minimum

        net.res_line_sc["ikss_ka"] = minmax(ppc["internal"]["branch_ikss_f"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1),
                                            ppc["internal"]["branch_ikss_t"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1))

        net.res_line_sc["ikss_from_ka"] = _get_branch_result_from_internal("branch_ikss_f", ppc, ppc_index, f, t)
        net.res_line_sc["ikss_from_degree"] = _get_branch_result_from_internal("branch_ikss_angle_f", ppc, ppc_index, f, t)

        net.res_line_sc["ikss_to_ka"] = _get_branch_result_from_internal("branch_ikss_t", ppc, ppc_index, f, t)
        net.res_line_sc["ikss_to_degree"] = _get_branch_result_from_internal("branch_ikss_angle_t", ppc, ppc_index, f, t)

        net.res_line_sc["p_from_mw"] = _get_branch_result_from_internal("branch_pkss_f", ppc, ppc_index, f, t)
        net.res_line_sc["q_from_mvar"] = _get_branch_result_from_internal("branch_qkss_f", ppc, ppc_index, f, t)

        net.res_line_sc["p_to_mw"] = _get_branch_result_from_internal("branch_pkss_t", ppc, ppc_index, f, t)
        net.res_line_sc["q_to_mvar"] = _get_branch_result_from_internal("branch_qkss_t", ppc, ppc_index, f, t)

        net.res_line_sc["vm_from_pu"] = _get_branch_result_from_internal("branch_vkss_f", ppc, ppc_index, f, t)
        net.res_line_sc["va_from_degree"] = _get_branch_result_from_internal("branch_vkss_angle_f", ppc, ppc_index, f, t)

        net.res_line_sc["vm_to_pu"] = _get_branch_result_from_internal("branch_vkss_t", ppc, ppc_index, f, t)
        net.res_line_sc["va_to_degree"] = _get_branch_result_from_internal("branch_vkss_angle_t", ppc, ppc_index, f, t)

        if net._options["ip"]:
            net.res_line_sc["ip_ka"] = minmax(ppc["internal"]["branch_ip_f"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1),
                                              ppc["internal"]["branch_ip_t"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1))
        if net._options["ith"]:
            net.res_line_sc["ith_ka"] = minmax(ppc["internal"]["branch_ith_f"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1),
                                               ppc["internal"]["branch_ith_t"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1))

def _get_switch_all_results(net, ppc, bus):
    case = net._options["case"]

    bus, ppc_index = _get_bus_ppc_idx_for_br_all_results(net, ppc, bus)
    branch_lookup = net._pd2ppc_lookups["branch"]

    multindex = pd.MultiIndex.from_product([net.res_switch_sc.index, bus], names=['switch','bus'])
    net.res_switch_sc = net.res_switch_sc.reindex(multindex)

    if "switch" in branch_lookup:
        f, t = branch_lookup["switch"]
        minmax = np.maximum if case == "max" else np.minimum

        net.res_switch_sc["ikss_ka"] = minmax(ppc["internal"]["branch_ikss_f"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1),
                                            ppc["internal"]["branch_ikss_t"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1))
        if net._options["ip"]:
            net.res_switch_sc["ip_ka"] = minmax(ppc["internal"]["branch_ip_f"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1),
                                              ppc["internal"]["branch_ip_t"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1))
        if net._options["ith"]:
            net.res_switch_sc["ith_ka"] = minmax(ppc["internal"]["branch_ith_f"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1),
                                               ppc["internal"]["branch_ith_t"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1))
    
    
def _get_trafo_results(net, ppc):
    branch_lookup = net._pd2ppc_lookups["branch"]
    if "trafo" in branch_lookup:
        f, t = branch_lookup["trafo"]
        net.res_trafo_sc["ikss_hv_ka"] = ppc["branch"][f:t, IKSS_F].real
        net.res_trafo_sc["ikss_hv_degree"] = ppc["branch"][f:t, IKSS_ANGLE_F].real
        net.res_trafo_sc["ikss_lv_ka"] = ppc["branch"][f:t, IKSS_T].real
        net.res_trafo_sc["ikss_lv_degree"] = ppc["branch"][f:t, IKSS_ANGLE_T].real

        # adding columns for new calculated VPQ
        net.res_trafo_sc["p_hv_mw"] = ppc["branch"][f:t, PKSS_F].real
        net.res_trafo_sc["q_hv_mvar"] = ppc["branch"][f:t, QKSS_F].real

        net.res_trafo_sc["p_lv_mw"] = ppc["branch"][f:t, PKSS_T].real
        net.res_trafo_sc["q_lv_mvar"] = ppc["branch"][f:t, QKSS_T].real

        net.res_trafo_sc["vm_hv_pu"] = ppc["branch"][f:t, VKSS_MAGN_F].real
        net.res_trafo_sc["va_hv_degree"] = ppc["branch"][f:t, VKSS_ANGLE_F].real

        net.res_trafo_sc["vm_lv_pu"] = ppc["branch"][f:t, VKSS_MAGN_T].real
        net.res_trafo_sc["va_lv_degree"] = ppc["branch"][f:t, VKSS_ANGLE_T].real


def _get_trafo_all_results(net, ppc, bus):
    bus, ppc_index = _get_bus_ppc_idx_for_br_all_results(net, ppc, bus)
    branch_lookup = net._pd2ppc_lookups["branch"]

    multindex = pd.MultiIndex.from_product([net.res_trafo_sc.index, bus], names=['trafo', 'bus'])
    net.res_trafo_sc = net.res_trafo_sc.reindex(multindex)

    if "trafo" in branch_lookup:
        f, t = branch_lookup["trafo"]
        net.res_trafo_sc["ikss_hv_ka"] = ppc["internal"]["branch_ikss_f"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1)
        net.res_trafo_sc["ikss_lv_ka"] = ppc["internal"]["branch_ikss_t"].iloc[f:t,:].loc[:, ppc_index].values.real.reshape(-1, 1)


def _get_trafo3w_results(net, ppc):
    branch_lookup = net._pd2ppc_lookups["branch"]
    if "trafo3w" in branch_lookup:
        f, t = net._pd2ppc_lookups["branch"]["trafo3w"]
        hv = int(f + (t - f) / 3)
        mv = int(f + 2 * (t - f) / 3)
        lv = t
        net.res_trafo3w_sc["ikss_hv_ka"] = ppc["branch"][f:hv, IKSS_F].real
        net.res_trafo3w_sc["ikss_mv_ka"] = ppc["branch"][hv:mv, IKSS_T].real
        net.res_trafo3w_sc["ikss_lv_ka"] = ppc["branch"][mv:lv, IKSS_T].real


def _get_trafo3w_all_results(net, ppc, bus):
    bus, ppc_index = _get_bus_ppc_idx_for_br_all_results(net, ppc, bus)
    branch_lookup = net._pd2ppc_lookups["branch"]

    multindex = pd.MultiIndex.from_product([net.res_trafo3w_sc.index, bus], names=['trafo3w', 'bus'])
    net.res_trafo3w_sc = net.res_trafo3w_sc.reindex(multindex)

    if "trafo3w" in branch_lookup:
        f, t = branch_lookup["trafo3w"]
        hv = int(f + (t - f) / 3)
        mv = int(f + 2 * (t - f) / 3)
        lv = t
        net.res_trafo3w_sc["ikss_hv_ka"] = ppc["internal"]["branch_ikss_f"].iloc[f:hv,:].loc[:, ppc_index].values.real.reshape(-1, 1)
        net.res_trafo3w_sc["ikss_mv_ka"] = ppc["internal"]["branch_ikss_t"].iloc[hv:mv, :].loc[:, ppc_index].values.real.reshape(-1, 1)
        net.res_trafo3w_sc["ikss_lv_ka"] = ppc["internal"]["branch_ikss_t"].iloc[mv:lv, :].loc[:, ppc_index].values.real.reshape(-1, 1)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


from pandapower.networks.kerber_networks import _create_branched_loads_network, _create_branch_network

# The Kerber extreme networks complement the typical Kerber networks.
# Categories:
# I. Kerber networks with extreme lines
# II. Kerber networks with extreme lines and high loaded transformer


# --- Extreme Landnetze

# I.  Kerber Landnetz mit extremen Netzstrahlen (Freileitung):
def kb_extrem_landnetz_freileitung(n_lines=26, l_lines_in_km=0.012, std_type="NFA2X 4x70",
                                   # trafotype="0.4 MVA 10/0.4 kV Yyn6 4 ASEA"
                                   trafotype="0.25 MVA 10/0.4 kV", p_load_mw=0.008,
                                   q_load_mvar=0, v_os=10.):
    pd_net = _create_branch_network(num_lines=[n_lines], len_lines=[l_lines_in_km],
                                    trafotype=trafotype, v_os=v_os,
                                    std_type=std_type, p_load_mw=p_load_mw, q_load_mvar=q_load_mvar)
    return pd_net


# I.  Kerber Landnetz mit extremen Netzstrahlen (Kabel):
def kb_extrem_landnetz_kabel(n_branch_1=26, l_lines_1_in_km=0.026, std_type="NAYY 4x150",
                             # trafotype="0.4 MVA 10/0.4 kV Yyn6 4 ASEA"
                             trafotype="0.25 MVA 10/0.4 kV", p_load_mw=.008,
                             q_load_mvar=0., length_branchout_line_1=0.018,
                             length_branchout_line_2=0.033, std_type_branchout_line="NAYY 4x50",
                             v_os=10.):
    num_lines = [n_branch_1]
    len_lines = [l_lines_1_in_km]
    pd_net = _create_branched_loads_network(trafotype=trafotype, v_os=v_os, num_lines=num_lines,
                                    len_lines=len_lines, std_type=std_type,
                                    p_load_mw=p_load_mw,
                                    q_load_mvar=q_load_mvar,
                                    length_branchout_line_1=length_branchout_line_1,
                                    length_branchout_line_2=length_branchout_line_2,
                                    std_type_branchout_line_1=std_type_branchout_line)
    return pd_net


# II.  Kerber Landnetz mit extremen Netzstrahlen (Freileitung) und hoch ausgelastetem Transformator:
def kb_extrem_landnetz_freileitung_trafo(n_branch_1=26, n_branch_2=1, l_lines_1_in_km=0.012,
                                         l_lines_2_in_km=0.036, std_type="NFA2X 4x70",
                                         trafotype="0.1 MVA 10/0.4 kV",
                                         p_load_mw=.008, q_load_mvar=0,
                                         v_os=10.):
    # num_lines = [26, 1] <- Solution with _add_lines_with_branched_loads, if branches = 0
    # len_lines = [0.012, 0.036]
    num_lines = [n_branch_1, n_branch_2]
    len_lines = [l_lines_1_in_km, l_lines_2_in_km]
    pd_net = _create_branch_network(num_lines=num_lines, len_lines=len_lines, trafotype=trafotype, v_os=v_os,
                             std_type=std_type, p_load_mw=p_load_mw,
                             q_load_mvar=q_load_mvar)

    return pd_net


# II. Kerber Landnetz mit extremen Netzstrahlen (Kabel) und hoch ausgelastetem Transformator:
def kb_extrem_landnetz_kabel_trafo(n_branch_1=26, n_branch_2=1, l_lines_1_in_km=0.026,
                                   l_lines_2_in_km=0.078, std_type="NAYY 4x150",
                                   trafotype="0.1 MVA 10/0.4 kV", p_load_mw=.008,
                                   q_load_mvar=0., length_branchout_line_1=0.018,
                                   length_branchout_line_2=0.033,
                                   std_type_branchout_line="NAYY 4x50",
                                   v_os=10.):
    num_lines = [n_branch_1, n_branch_2]
    len_lines = [l_lines_1_in_km, l_lines_2_in_km]
    pd_net = _create_branched_loads_network(len_lines=len_lines, num_lines=num_lines,
                                   std_type=std_type, p_load_mw=p_load_mw, v_os=v_os,
                                   q_load_mvar=q_load_mvar, trafotype=trafotype,
                                   length_branchout_line_1=length_branchout_line_1,
                                   length_branchout_line_2=length_branchout_line_2,
                                   std_type_branchout_line_1=std_type_branchout_line)
    return pd_net


# --- Extreme Dorfnetze

# I. Kerber Dorfnetz mit extremen Netzstrahlen:
def kb_extrem_dorfnetz(std_type="NAYY 4x150", trafotype="0.4 MVA 10/0.4 kV",
                       p_load_mw=.006, q_load_mvar=0., length_branchout_line_1=0.015,
                       length_branchout_line_2=0.031, std_type_branchout_line="NAYY 4x50",
                       v_os=10.):
    num_lines = [28, 16, 9, 4, 1]
    len_lines = [0.021, 0.029, 0.040, 0.064, 0.102]
    pd_net = _create_branched_loads_network(len_lines=len_lines, num_lines=num_lines,
                                   std_type=std_type, p_load_mw=p_load_mw, v_os=v_os,
                                   q_load_mvar=q_load_mvar, trafotype=trafotype,
                                   length_branchout_line_1=length_branchout_line_1,
                                   length_branchout_line_2=length_branchout_line_2,
                                   std_type_branchout_line_1=std_type_branchout_line)
    return pd_net


# II. Kerber Dorfnetz mit extremen Netzstrahlen und hoch ausgelastetem Transformator:
def kb_extrem_dorfnetz_trafo(std_type="NAYY 4x150", trafotype="0.25 MVA 10/0.4 kV",
                             # trafotype="0.4 MVA 10/0.4 kV Yyn6 4 ASEA",
                             p_load_mw=0.006, q_load_mvar=0., length_branchout_line_1=0.015,
                             length_branchout_line_2=0.031, std_type_branchout_line="NAYY 4x50",
                             v_os=10.):
    num_lines = [28, 28, 16, 12, 12, 9, 7, 4, 1]
    len_lines = [0.021, 0.021, 0.029, 0.032, 0.032, 0.040, 0.043, 0.064, 0.102]
    pd_net = _create_branched_loads_network(len_lines=len_lines, num_lines=num_lines,
                                   std_type=std_type, p_load_mw=p_load_mw, v_os=v_os,
                                   q_load_mvar=q_load_mvar, trafotype=trafotype,
                                   length_branchout_line_1=length_branchout_line_1,
                                   length_branchout_line_2=length_branchout_line_2,
                                   std_type_branchout_line_1=std_type_branchout_line)
    return pd_net


# --- Extreme Vorstadtnetze

# I. Kerber Vorstadtnetz mit extremen Netzstrahlen (Typ 1):
def kb_extrem_vorstadtnetz_1(std_type="NAYY 4x150", p_load_mw=0.002, q_load_mvar=0.,
                             trafotype="0.63 MVA 10/0.4 kV", v_os=10.):

    num_lines = [69, 32, 19, 14, 10, 1]
    len_lines = [0.006, 0.011, 0.017, 0.021, 0.025, 0.068]
    pd_net = _create_branched_loads_network(len_lines=len_lines, num_lines=num_lines,
                                   std_type=std_type, p_load_mw=p_load_mw, v_os=v_os,
                                   q_load_mvar=q_load_mvar, trafotype=trafotype,
                                   length_branchout_line_1=0.011,
                                   std_type_branchout_line_1="NAYY 4x50",
                                   std_type_branchout_line_2="NYY 4x35",
                                   prob_branchout_line_1=0.5)
    return pd_net


# I. Kerber Vorstadtnetz mit extremen Netzstrahlen (Typ 2):
def kb_extrem_vorstadtnetz_2(std_type="NAYY 4x185", p_load_mw=0.002, q_load_mvar=0.,
                             trafotype="0.63 MVA 10/0.4 kV", v_os=10.):
    num_lines = [61, 32, 20, 15, 11, 5, 1]
    len_lines = [0.01, 0.014, 0.02, 0.023, 0.026, 0.05, 0.085]
    pd_net = _create_branched_loads_network(len_lines=len_lines, num_lines=num_lines,
                                   std_type=std_type, p_load_mw=p_load_mw, v_os=v_os,
                                   q_load_mvar=q_load_mvar, trafotype=trafotype,
                                   length_branchout_line_1=0.011,
                                   std_type_branchout_line_1="NAYY 4x50",
                                   std_type_branchout_line_2="NYY 4x35",
                                   prob_branchout_line_1=0.5)
    return pd_net


# II. Kerber Vorstadtnetz mit extremen Netzstrahlen und hoch ausgeslatetem Transformator:
def kb_extrem_vorstadtnetz_trafo_1(std_type="NAYY 4x150", p_load_mw=0.002, q_load_mvar=0.,
                                   trafotype="0.25 MVA 10/0.4 kV", v_os=10.):
    num_lines = [69, 32, 32, 19, 14, 10, 10, 4, 1]
    len_lines = [0.006, 0.011, 0.011, 0.017, 0.021, 0.025, 0.025, 0.06, 0.68]
    pd_net = _create_branched_loads_network(len_lines=len_lines, num_lines=num_lines,
                                   std_type=std_type, p_load_mw=p_load_mw, v_os=v_os,
                                   q_load_mvar=q_load_mvar, trafotype=trafotype,
                                   length_branchout_line_1=0.011,
                                   std_type_branchout_line_1="NAYY 4x50",
                                   std_type_branchout_line_2="NYY 4x35",
                                   prob_branchout_line_1=0.5)
    return pd_net


# II. Kerber Vorstadtnetz mit extremen Netzstrahlen und hoch ausgeslatetem Transformator:
def kb_extrem_vorstadtnetz_trafo_2(std_type="NAYY 4x185", p_load_mw=0.002, q_load_mvar=0.,
                                   trafotype="0.25 MVA 10/0.4 kV", v_os=10.):
    num_lines = [61, 32, 32, 20, 15, 15, 11, 5, 1]
    len_lines = [0.01, 0.014, 0.014, 0.02, 0.023, 0.023, 0.026, 0.05, 0.085]
    pd_net = _create_branched_loads_network(len_lines=len_lines, num_lines=num_lines,
                                   std_type=std_type, p_load_mw=p_load_mw, v_os=v_os,
                                   q_load_mvar=q_load_mvar, trafotype=trafotype,
                                   length_branchout_line_1=0.011,
                                   std_type_branchout_line_1="NAYY 4x50",
                                   std_type_branchout_line_2="NYY 4x35",
                                   prob_branchout_line_1=0.5)
    return pd_net


try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)

def create_network_dict(app, flag_graphics='GPS'):
    # elements to be exported from PowerFactory
    set_object_extentions = {
        # node elements:
        'ElmTerm',
        'ElmXnet',
        'ElmLod',
        'ElmLodlv',
        'ElmLodlvp',
        'ElmLodmv',
        'ElmGenstat',
        'ElmPvsys',
        'ElmSym',
        'ElmAsm',
        'ElmShnt',
        'ElmVac',
        'ElmSvs',
        'ElmVsc',
        'ElmVscmono',


        # branch elements:
        'ElmLne',
        'ElmCoup',
        'RelFuse',
        'ElmZpu',
        'ElmSind',
        'ElmScap',
        'StaSwitch',
        'ElmTr2',
        'ElmTr3',

        # we don't gather types anymore, because they are not collected for elements that are out
        #  of service
        #  types:
        # 'TypLne',
        # 'TypTr2',
        # 'TypTr3'

        # controllers:
        'ElmStactrl'
    }

    # here define all element types that have to be configured to have MW as power values
    elm_units = {
        'ElmLod': ['W', 'var', 'VA'],
        'ElmLodlv': ['W', 'var', 'VA'],
        'ElmLodlvp': ['W', 'var', 'VA'],
        'ElmLodmv': ['W', 'var', 'VA'],
        'ElmGenstat': ['W', 'var', 'VA'],
        'ElmPvsys': ['W', 'var', 'VA'],
        'ElmXnet': ['W', 'var', 'VA'],
        'ElmSym': ['W', 'var', 'VA'],
        'ElmSvs': ['W', 'var', 'VA'],
        'ElmAsm': ['W', 'var', 'VA'],
        'ElmShnt': ['W', 'var', 'VA'],
        'ElmZpu': ['W', 'var', 'VA'],
        'ElmSind': ['W', 'var', 'VA', 'V'],
        'ElmScap': ['W', 'var', 'VA', 'V'],
        'ElmVac': ['W', 'var', 'VA'],
        'ElmTr2': ['W', 'var'],
        'ElmTr3': ['W', 'var'],
        'TypTr2': ['W', 'var', 'VA'],
        'TypTr3': ['W', 'var', 'VA'],
        'TypLne': ['A', 'm']
    }

    # make all values in MW
    logger.info('applying unit settings')
    # echo_on(app)
    # apply_unit_settings(app, elm_units, 'M')
    # echo_off(app)

    dict_net = {}

    grid = app.GetSummaryGrid()
    dict_net['ElmNet'] = grid
    dict_net['global_parameters'] = get_global_parameters(app)

    logger.info('collecting network elements')
    for obj in set_object_extentions:
        if obj == 'ElmTerm':
            dict_net[obj] = app.GetCalcRelevantObjects(obj, 1, 0, 1)
        else:
            dict_net[obj] = app.GetCalcRelevantObjects(obj)

    if flag_graphics not in ['GPS', 'no geodata']:
        logger.info('gathering graphic objects')
        dia_name = flag_graphics
        gather_graphic_objects(app, dict_net, dia_name)

    dict_net['lvp_params'] = get_lvp_params(app)

    return dict_net


def get_lvp_params(app):
    """
    Gather parameters for partial LV loads (loads that model an LV customer, the load elements are
    stored in a line element without having own buses in PowerFactory, so the lines in pandapower
    must be split at those points and new buses must be created).
    """
    com_ldf = app.GetFromStudyCase('ComLdf')

    lvp_params = {
        'iopt_sim': com_ldf.iopt_sim,
        'scPnight': com_ldf.scPnight,
        'Sfix': com_ldf.Sfix,
        'cosfix': com_ldf.cosfix,
        'Svar': com_ldf.Svar,
        'cosvar': com_ldf.cosvar,
        'ginf': com_ldf.ginf,
        'i_volt': com_ldf.i_volt,
        'hunting_limit': com_ldf.maxTapTrans
    }

    return lvp_params


def get_global_parameters(app):
    prj = app.GetActiveProject()
    settings = prj.pPrjSettings
    base_sn_mva = settings.Sbase

    # global load and generation scaling
    com_ldf = app.GetFromStudyCase('ComLdf')
    global_load_scaling = com_ldf.scLoadFac * 1e-2
    global_generation_scaling = com_ldf.scGenFac * 1e-2
    global_motor_scaling = com_ldf.scMotFac * 1e-2

    global_parameters = {
        'base_sn_mva': base_sn_mva,
        'global_load_scaling': global_load_scaling,
        'global_generation_scaling': global_generation_scaling,
        'global_motor_scaling': global_motor_scaling,
        'iopt_tem': com_ldf.iopt_tem,  # calculate load flow at 20 °C or at max. temperature
        'global_load_voltage_dependency': com_ldf.iopt_pq  # consider voltage dependency of loads
    }
    return global_parameters


def gather_graphic_objects(app, dict_net, dia_name):
    dia_folder = app.GetProjectFolder('dia')
    dia_grid = dia_folder.GetContents(dia_name + '.IntGrfnet')
    grf_objs = []
    for d in dia_grid:
        grf_objs.extend(d.GetContents('*.IntGrf'))
    logger.debug('collected graphic objects: %s' % grf_objs)
    dict_net['graphics'] = {grf.pDataObj: grf for grf in grf_objs}


def apply_unit_settings(app, elm_units, exponent):
    prj = app.GetActiveProject()
    # don't forget to define the elements to apply the unit settings to!
    setup_project_power_exponent(prj, exponent)

    for key, val in elm_units.items():
        for unit in val:
            setup_unit_exponents(prj, key, unit, exponent)

    # special case for c_nf_per_km in nf and I in kA:
    for elm in ['TypLne', 'ElmLne', 'ElmLnesec']:
        setup_unit_exponents(prj, elm, 'F/km', 'n')
        setup_unit_exponents(prj, elm, 'F', 'n')
        setup_unit_exponents(prj, elm, 'A', 'k')

    # to make PowerFactory apply the new settings
    prj.Deactivate()
    prj.Activate()


def run_load_flow(app, scale_feeder_loads=False, load_scaling=None, gen_scaling=None,
                  motor_scaling=None):
    """
    :param app: PowerFactory Application object
    :param scale_feeder_loads: if loads have to be scaled according to the feeder scaling factor
    :param motor_scaling: Load flow parameter in PowerFactory ("Load/Generation scaling")
    :param gen_scaling: Load flow parameter in PowerFactory ("Load/Generation scaling")
    :param load_scaling: Load flow parameter in PowerFactory ("Load/Generation scaling")
    :return: None
    """

    com_ldf = app.GetFromStudyCase('ComLdf')

    # com_ldf.iopt_net = 0
    # com_ldf.iopt_at = 1
    # com_ldf.iopt_pq = 0
    if scale_feeder_loads:
        logger.debug('scale_feeder_loads is True')
        com_ldf.iopt_fls = 1

    # com_ldf.errlf = 0.001
    # com_ldf.erreq = 0.01

    if com_ldf.iopt_sim == 1:
        logger.warning(f'Calculation method probabilistic loadflow of lv-loads is activated!'
                       f' The validation will not succeed.')
    if load_scaling is not None:
        logger.debug('scaling loads at %.2f' % load_scaling)
        com_ldf.scLoadFac = load_scaling
    if gen_scaling is not None:
        logger.debug('scaling generators at %.2f' % gen_scaling)
        com_ldf.scGenFac = gen_scaling
    if motor_scaling is not None:
        logger.debug('scaling motors at %.2f' % motor_scaling)
        com_ldf.scMotFac = motor_scaling

    logger.info('---------------------------------------------------------------------------------')
    logger.info('PowerFactory load flow settings:')
    # Active power regulation
    logger.info('Calculation method (AC balanced, AC unbalanced): %s' % com_ldf.iopt_net)
    logger.info(f'Calculation method probabilistic loadflow of lv-loads: {com_ldf.iopt_sim}')
    logger.info('Automatic tap adjustment of phase shifters: %s' % com_ldf.iPST_at)
    logger.info('Consider active power limits: %s' % com_ldf.iopt_plim)
    # Voltage and reactive power regulation
    logger.info('Automatic tap adjustment of transformers: %s' % com_ldf.iopt_at)
    logger.info('Automatic tap adjustment of shunts: %s' % com_ldf.iopt_asht)
    logger.info('Consider reactive power limits: %s' % com_ldf.iopt_lim)
    # Temperature dependency
    logger.info('Calculate at 20 °C/max. temperature: %s' % com_ldf.iopt_tem)
    # Load options
    logger.info('Consider voltage dependency of loads: %s' % com_ldf.iopt_pq)
    logger.info('Feeder load scaling: %s' % com_ldf.iopt_fls)
    logger.info('---------------------------------------------------------------------------------')

    res = com_ldf.Execute()
    # print("pf result", res)
    if res != 0:
        logger.error('Load flow failed due to divergence of %s loops' % ['inner', 'outer'][res - 1])
#        raise RuntimeError(
#            'Load flow failed due to divergence of %s loops' % ['inner', 'outer'][res - 1])
    return res


def setup_unit_exponents(prj, elm_class, unit, exponent):
    cdigexp = 'M' if unit in ['var', 'W', 'VA'] else 'u' if unit in ['F/km', 'F'] else 'k' if unit=='A' else ''
    logger.debug('setting unit exponents: %s, %s, %s, %s' % (elm_class, unit, cdigexp, exponent))
    settings_folder = prj.GetContents('*.SetFold')[0]
    units_folder = settings_folder.GetContents('*.IntUnit')

    if len(units_folder) == 0:
        units_folder = settings_folder.CreateObject('IntUnit', 'Units')
    else:
        units_folder = units_folder[0]

    object_name = elm_class + '-' + unit
    # delete old objects if there are any
    trash = units_folder.GetContents(object_name)
    for item in trash:
        item.Delete()

    unit_setting = units_folder.CreateObject('SetVariable', object_name)
    try:  # version-specific expectation at this point
        unit_setting.filtclass = elm_class if isinstance(elm_class, list) else [elm_class]
    except TypeError:
        unit_setting.filtclass = elm_class[0] if isinstance(elm_class, list) else elm_class
    unit_setting.digunit = unit
    # if unit in ['W', 'var', 'VA']:
    unit_setting.SetAttribute('cdigexp', cdigexp)
    unit_setting.cuserexp = exponent


def setup_project_power_exponent(prj, exponent):
    prj_settings = prj.pPrjSettings
    study_cases = prj.GetContents('*.IntCase', 1)

    prj_settings.cspqexp = exponent  # power MW
    prj_settings.clenexp = 'k'  # length km

    # for base Mva:
    setup_unit_exponents(prj, 'SetPrj', 'VA', 'M')

    for object in study_cases:
        object.cpowexp = exponent
        object.cpexpshc = exponent
        object.campexp = 'k'  # current kA



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import numpy as np
from scipy.sparse import csr_matrix


def calc_y_svc(x_control_degree, svc_x_l_ohm, svc_x_cvar_ohm, v_base_kv, baseMVA):
    """
    Computes the admittance of SVC (Static VAR Compensator) components based on their control angles
    and reactance values in ohms.

    Parameters:
    -----------
    x_control_degree : array-like
        Thyristor firing angles for the SVCs in degrees.
    svc_x_l_ohm : array-like
        Reactance of the SVC's inductor in ohms.
    svc_x_cvar_ohm : array-like
        Reactance of the SVC's capacitor in ohms.
    v_base_kv : array-like
        Base voltage levels in kV for corresponding SVC elements.
    baseMVA : float
        System's base power in MVA.

    Returns:
    --------
    array-like
        Computed admittance values of the SVCs in ohms.
    """
    x_control = np.deg2rad(x_control_degree)
    z_base_ohm = np.square(v_base_kv) / baseMVA
    svc_x_l_pu = svc_x_l_ohm / z_base_ohm
    svc_x_cvar_pu = svc_x_cvar_ohm / z_base_ohm
    y_svc = calc_y_svc_pu(x_control, svc_x_l_pu, svc_x_cvar_pu)
    y_svc *= z_base_ohm
    return y_svc


def calc_y_svc_pu(x_control, svc_x_l_pu, svc_x_cvar_pu):
    """
    Calculates the admittance of SVC elements (Static VAR Compensator) in per unit.

    Parameters:
    -----------
    x_control : array-like
        Thyristor firing angles in radians for the SVCs.
    svc_x_l_pu : array-like
        Reactance of the SVC's inductor in per-unit.
    svc_x_cvar_pu : array-like
        Reactance of the SVC's capacitor in per-unit.

    Returns:
    --------
    array-like
        Admittance values of the SVCs in per-unit.
    """
    y_svc = (2 * (np.pi - x_control) + np.sin(2 * x_control) + np.pi * svc_x_l_pu / svc_x_cvar_pu) / (np.pi * svc_x_l_pu)
    return y_svc


@np.errstate(all="raise")
def makeYbus_svc(Ybus, x_control_svc, svc_x_l_pu, svc_x_cvar_pu, svc_buses):
    """
    Constructs the SVC (Static VAR Compensator) admittance matrix.

    Parameters:
    -----------
    Ybus : csr_matrix
        The base admittance matrix of the system.
    x_control_svc : array-like
        Thyristor firing angles for the SVCs in degrees.
    svc_x_l_pu : array-like
        Reactance of the SVC's inductor in per-unit.
    svc_x_cvar_pu : array-like
        Reactance of the SVC's capacitor in per-unit.
    svc_buses : array-like
        Bus indices where SVCs are connected.

    Returns:
    --------
    csr_matrix
        The admittance matrix for the SVCs.
    """
    Y_SVC = -1j * calc_y_svc_pu(x_control_svc, svc_x_l_pu, svc_x_cvar_pu)
    Ybus_svc = csr_matrix((Y_SVC, (svc_buses, svc_buses)), shape=Ybus.shape, dtype=np.complex128)
    return Ybus_svc


def makeYbus_tcsc(Ybus, x_control_tcsc, tcsc_x_l_pu, tcsc_x_cvar_pu, tcsc_fb, tcsc_tb):
    """
    Constructs the TCSC (Thyristor Controlled Series Capacitor) admittance matrix.

    Parameters:
    -----------
    Ybus : csr_matrix
        The base admittance matrix of the system.
    x_control_tcsc : array-like
        Thyristor firing angles for the TCSCs in radians.
    tcsc_x_l_pu : array-like
        Reactance of the TCSC's inductor in per-unit.
    tcsc_x_cvar_pu : array-like
        Reactance of the TCSC's capacitor in per-unit.
    tcsc_fb : array-like
        From-bus indices where TCSCs are connected.
    tcsc_tb : array-like
        To-bus indices where TCSCs are connected.

    Returns:
    --------
    csr_matrix
        The updated admittance matrix considering the TCSCs.
    """
    Y_TCSC = -1j * calc_y_svc_pu(x_control_tcsc, tcsc_x_l_pu, tcsc_x_cvar_pu)
    Ybus_tcsc = make_Ybus_facts(tcsc_fb, tcsc_tb, Y_TCSC, Ybus.shape[0])
    return Ybus_tcsc


def makeYft_tcsc(Ybus_tcsc, tcsc_fb, tcsc_tb, x_control_tcsc, tcsc_x_l_pu, tcsc_x_cvar_pu):
    """
    Constructs the 'from' and 'to' admittance matrices for the TCSC (Thyristor Controlled Series Capacitor).

    Parameters:
    -----------
    Ybus_tcsc : csr_matrix
        The admittance matrix considering the TCSCs.
    x_control_tcsc : array-like
        Thyristor firing angles for the TCSCs in radians.
    tcsc_x_l_pu : array-like
        Reactance of the TCSC's inductor in per-unit.
    tcsc_x_cvar_pu : array-like
        Reactance of the TCSC's capacitor in per-unit.
    tcsc_fb : array-like
        From-bus indices where TCSCs are connected.
    tcsc_tb : array-like
        To-bus indices where TCSCs are connected.

    Returns:
    --------
    tuple of csr_matrix
        - Yf : csr_matrix
            The 'from' admittance matrix for the TCSCs.
        - Yt : csr_matrix
            The 'to' admittance matrix for the TCSCs.
    """
    Y_TCSC = -1j * calc_y_svc_pu(x_control_tcsc, tcsc_x_l_pu, tcsc_x_cvar_pu)
    Yf, Yt = make_Yft_facts(tcsc_fb, tcsc_tb, Y_TCSC, Ybus_tcsc.shape[0])
    return Yf, Yt


def makeYbus_ssc_vsc(Ybus, internal_y_pu, fb, tb, controllable):
    """
    Constructs the admittance matrices for SSCs (Static Synchronous Compensators)
    and VSCs (Voltage Source Converters).

    Parameters:
    -----------
    Ybus : csr_matrix
        The base admittance matrix.
    internal_y_pu : array-like
        Internal transformer admittance values in per-unit.
    fb : array-like
        From bus indices.
    tb : array-like
        To bus indices.
    controllable : array-like
        Boolean flags indicating if elements are controllable.

    Returns:
    --------
    tuple of csr_matrix
        The not controllable, controllable, and combined admittance matrices.
    """
    Ybus_matrices = []

    for flag in [~controllable, controllable]:
        if not np.any(flag):
            Ybus_matrices.append(csr_matrix(Ybus.shape, dtype=np.complex128))
            continue
        Ybus_flag = make_Ybus_facts(fb[flag], tb[flag], internal_y_pu[flag], Ybus.shape[0])
        Ybus_matrices.append(Ybus_flag)

    Ybus_not_controllable, Ybus_controllable = Ybus_matrices[0], Ybus_matrices[1]

    return Ybus_not_controllable, Ybus_controllable, Ybus_not_controllable + Ybus_controllable


def make_Ybus_facts(from_bus, to_bus, y_pu, n, ysf_pu=0, yst_pu=0, dtype=np.complex128):
    """
    Construct the bus admittance matrix with an added FACTS device for a power grid.

    This function creates an additional Ybus matrix, which models the connectivity of
    additional FACTS (Flexible AC Transmission System) devices in the power grid.

    Parameters:
    - from_bus (array-like): Array of source bus indices for each element.
    - to_bus (array-like): Array of destination bus indices for each element.
    - y_pu (array-like): Array of per-unit admittances for each element.
    - n (int): Total number of buses in the grid.

    Returns:
    - csr_matrix: A compressed sparse row representation of the bus admittance matrix
                  incorporating FACTS.
    """
    # Combine diagonal elements with off-diagonal elements
    # Concatenate row and column indices, as well as data
    # Explanation: first pair is for diagonal data, second pair is for off-diagonal data
    # At creation of the csr_matrix, data elements at the coinciding indices are added
    # This feature is useful when you have repeated indices, and you want their values to be aggregated
    row_indices = np.concatenate([from_bus, to_bus, from_bus, to_bus])
    col_indices = np.concatenate([from_bus, to_bus, to_bus, from_bus])
    data = np.concatenate([y_pu + ysf_pu, y_pu + yst_pu, -y_pu, -y_pu])

    # Create and return the Ybus matrix using the compressed sparse row format
    Ybus_facts = csr_matrix((data, (row_indices, col_indices)), shape=(n, n), dtype=dtype)
    return Ybus_facts


@np.errstate(all="raise")
def make_Yft_facts(from_bus, to_bus, y_pu, n, ysf_pu=0, yst_pu=0):
    """
    Construct the Yf and Yt admittance matrices for branches with FACTS devices.

    This function creates the Yf and Yt matrices, which allow to compute
    the currents injected at the "from" and "to" ends of each branch,
    respectively. The matrices represent the FACTS (Flexible AC Transmission
    System) devices in the grid.

    Parameters:
    - from_bus (array-like): Array of source bus indices for each branch.
    - to_bus (array-like): Array of destination bus indices for each branch.
    - y_pu (array-like): Array of per-unit admittances for each branch.
    - n (int): Total number of buses in the grid.

    Returns:
    - Yf (csr_matrix): Admittance matrix for the "from" end of each branch.
    - Yt (csr_matrix): Admittance matrix for the "to" end of each branch.
    """

    # Number of lines (or branches)
    nl = len(from_bus)
    rows = np.arange(nl)

    # Common row indices for Yf and Yt matrices
    row_indices = np.concatenate([rows, rows])
    # Common column indices for Yf and Yt matrices
    col_indices = np.concatenate([from_bus, to_bus])

    # Construct Yf and Yt matrices using the CSR format
    Yf = csr_matrix((np.concatenate([y_pu + ysf_pu, -y_pu]), (row_indices, col_indices)), shape=(nl, n))
    Yt = csr_matrix((np.concatenate([-y_pu, y_pu + yst_pu]), (row_indices, col_indices)), shape=(nl, n))

    return Yf, Yt



import logging
import time

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.synchronousMachinesCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class SynchronousMachinesCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_synchronous_machines_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting SynchronousMachines.")
        eqssh_synchronous_machines = self._prepare_synchronous_machines_cim16()

        # Add synchronous machine characteristics table and id_characteristics_table
        eqssh_synchronous_machines = self._create_gen_characteristics_table(eqssh_synchronous_machines)

        # convert the SynchronousMachines with voltage control to gens
        eqssh_sm_gens = eqssh_synchronous_machines.loc[(eqssh_synchronous_machines['mode'] == 'voltage') &
                                                       (eqssh_synchronous_machines['enabled'])]
        self.cimConverter.copy_to_pp('gen', eqssh_sm_gens)
        # now deal with the pq generators
        eqssh_synchronous_machines = eqssh_synchronous_machines.loc[(eqssh_synchronous_machines['mode'] != 'voltage') |
                                                                    (~eqssh_synchronous_machines['enabled'])]
        self.cimConverter.copy_to_pp('sgen', eqssh_synchronous_machines)
        self.logger.info("Created %s gens and %s sgens in %ss." %
                         (eqssh_sm_gens.index.size, eqssh_synchronous_machines.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s gens and %s sgens from SynchronousMachines in %ss." %
                    (eqssh_sm_gens.index.size, eqssh_synchronous_machines.index.size, time.time() - time_start)))

    def _prepare_synchronous_machines_cim16(self) -> pd.DataFrame:
        eq_generating_units = self.cimConverter.cim['eq']['GeneratingUnit'][
            ['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']]
        # a column for the type of the static generator in pandapower
        eq_generating_units['type'] = 'GeneratingUnit'
        eq_generating_units = pd.concat([eq_generating_units, self.cimConverter.cim['eq']['WindGeneratingUnit']],
                                        sort=False)
        eq_generating_units['type'] = eq_generating_units['type'].fillna('WP')
        eq_generating_units = pd.concat([eq_generating_units, self.cimConverter.cim['eq']['HydroGeneratingUnit']],
                                        sort=False)
        eq_generating_units['type'] = eq_generating_units['type'].fillna('Hydro')
        eq_generating_units = pd.concat([eq_generating_units, self.cimConverter.cim['eq']['SolarGeneratingUnit']],
                                        sort=False)
        eq_generating_units['type'] = eq_generating_units['type'].fillna('PV')
        eq_generating_units = pd.concat([eq_generating_units, self.cimConverter.cim['eq']['ThermalGeneratingUnit']],
                                        sort=False)
        eq_generating_units['type'] = eq_generating_units['type'].fillna('Thermal')
        eq_generating_units = pd.concat([eq_generating_units, self.cimConverter.cim['eq']['NuclearGeneratingUnit']],
                                        sort=False)
        eq_generating_units['type'] = eq_generating_units['type'].fillna('Nuclear')
        eq_generating_units = eq_generating_units.rename(columns={'rdfId': 'GeneratingUnit'})

        if 'sc' in self.cimConverter.cim.keys():
            synchronous_machines = self.cimConverter.merge_eq_other_profiles(
                ['ssh', 'sc'], 'SynchronousMachine', add_cim_type_column=True)
        else:
            synchronous_machines = self.cimConverter.merge_eq_ssh_profile('SynchronousMachine',
                                                                          add_cim_type_column=True)

        if 'type' in synchronous_machines.columns:
            synchronous_machines = synchronous_machines.drop(columns=['type'])
        if 'EquipmentContainer' in synchronous_machines.columns:
            synchronous_machines = synchronous_machines.drop(columns=['EquipmentContainer'])
        synchronous_machines = pd.merge(synchronous_machines, eq_generating_units,
                                        how='left', on='GeneratingUnit')
        eqssh_reg_control = self.cimConverter.merge_eq_ssh_profile('RegulatingControl')[
            ['rdfId', 'mode', 'enabled', 'targetValue', 'Terminal']]
        # if voltage is not on connectivity node, topological node will be used
        eqtp_terminals = pd.merge(self.cimConverter.cim['eq']['Terminal'], self.cimConverter.cim['tp']['Terminal'],
                                  how='left', on='rdfId')
        eqtp_terminals.ConnectivityNode = eqtp_terminals.ConnectivityNode.fillna(eqtp_terminals.TopologicalNode)
        eqtp_terminals = eqtp_terminals[['rdfId', 'ConnectivityNode']].rename(
            columns={'rdfId': 'Terminal', 'ConnectivityNode': 'reg_control_cnode'})
        eqssh_reg_control = pd.merge(eqssh_reg_control, eqtp_terminals, how='left', on='Terminal')
        eqssh_reg_control = eqssh_reg_control.drop(columns=['Terminal'])
        # add the voltage from the bus
        eqssh_reg_control = pd.merge(eqssh_reg_control, self.cimConverter.net.bus[['vn_kv', sc['o_id']]].rename(
            columns={sc['o_id']: 'reg_control_cnode'}), how='left', on='reg_control_cnode')
        # merge with RegulatingControl to check if it is a voltage controlled generator
        eqssh_reg_control = eqssh_reg_control.loc[eqssh_reg_control['mode'] == 'voltage']
        synchronous_machines = pd.merge(
            synchronous_machines, eqssh_reg_control.rename(columns={'rdfId': 'RegulatingControl'}),
            how='left', on='RegulatingControl')
        synchronous_machines["controllable"] = synchronous_machines['controlEnabled'] & synchronous_machines['enabled']
        synchronous_machines = pd.merge(synchronous_machines, self.cimConverter.bus_merge, how='left', on='rdfId')
        synchronous_machines = synchronous_machines.drop_duplicates(['rdfId'], keep='first')
        synchronous_machines['vm_pu'] = synchronous_machines.targetValue / synchronous_machines.vn_kv
        synchronous_machines['vm_pu'] = synchronous_machines['vm_pu'].fillna(1.)
        synchronous_machines = synchronous_machines.rename(columns={'vn_kv': 'bus_voltage'})
        synchronous_machines['slack'] = False
        # set the slack = True for gens with highest prio
        # get the highest prio from SynchronousMachines
        sync_ref_prio_min = synchronous_machines.loc[
            (synchronous_machines['referencePriority'] > 0) & (synchronous_machines['enabled']),
            'referencePriority'].min()
        # get the highest prio from ExternalNetworkInjection and check if the slack is an ExternalNetworkInjection
        enis = self.cimConverter.merge_eq_ssh_profile('ExternalNetworkInjection')
        regulation_controllers = self.cimConverter.merge_eq_ssh_profile('RegulatingControl')
        regulation_controllers = regulation_controllers.loc[regulation_controllers['mode'] == 'voltage']
        regulation_controllers = regulation_controllers[['rdfId', 'targetValue', 'enabled']]
        regulation_controllers = regulation_controllers.rename(columns={'rdfId': 'RegulatingControl'})
        enis = pd.merge(enis, regulation_controllers, how='left', on='RegulatingControl')

        eni_ref_prio_min = enis.loc[(enis['referencePriority'] > 0) & (enis['enabled']), 'referencePriority'].min()
        if pd.isna(sync_ref_prio_min):
            ref_prio_min = eni_ref_prio_min
        elif pd.isna(eni_ref_prio_min):
            ref_prio_min = sync_ref_prio_min
        else:
            ref_prio_min = min(eni_ref_prio_min, sync_ref_prio_min)

        synchronous_machines.loc[synchronous_machines['referencePriority'] == ref_prio_min, 'slack'] = True
        synchronous_machines['p_mw'] = -synchronous_machines['p']
        synchronous_machines['q_mvar'] = -synchronous_machines['q']
        synchronous_machines['current_source'] = True
        synchronous_machines['sn_mva'] = \
            synchronous_machines['ratedS'].fillna(synchronous_machines['nominalP'])
        # Convert governorSCD unit from percent to MW/Hz
        synchronous_machines['governorSCD'] = synchronous_machines['governorSCD'] * synchronous_machines['nominalP'] / 50 / 100
        # SC data
        synchronous_machines['vn_kv'] = synchronous_machines['ratedU'][:]
        synchronous_machines['rdss_ohm'] = \
            synchronous_machines['r2'] * \
            (synchronous_machines['ratedU'] ** 2 / synchronous_machines['ratedS'])
        synchronous_machines['xdss_pu'] = synchronous_machines['x2'][:]
        synchronous_machines['voltageRegulationRange'] = synchronous_machines['voltageRegulationRange'].fillna(0.)
        synchronous_machines['pg_percent'] = synchronous_machines['voltageRegulationRange']
        synchronous_machines['k'] = (synchronous_machines['ratedS'] * 1e3 / synchronous_machines[
            'ratedU']) / (synchronous_machines['ratedU'] / (
                3 ** .5 * (synchronous_machines['r2'] ** 2 + synchronous_machines['x2'] ** 2)))
        synchronous_machines['rx'] = synchronous_machines['r2'] / synchronous_machines['x2']
        synchronous_machines['scaling'] = 1.
        synchronous_machines['generator_type'] = 'current_source'
        synchronous_machines.loc[synchronous_machines['referencePriority'] == 0, 'referencePriority'] = float('NaN')
        synchronous_machines['referencePriority'] = synchronous_machines['referencePriority'].astype(float)
        if 'inService' in synchronous_machines.columns:
            synchronous_machines['connected'] = (synchronous_machines['connected'] & synchronous_machines['inService'])
        synchronous_machines = synchronous_machines.rename(columns={
            'rdfId_Terminal': sc['t'], 'rdfId': sc['o_id'], 'connected': 'in_service', 'index_bus': 'bus',
            'minOperatingP': 'min_p_mw', 'maxOperatingP': 'max_p_mw', 'minQ': 'min_q_mvar', 'maxQ': 'max_q_mvar',
            'ratedPowerFactor': 'cos_phi', 'referencePriority': 'slack_weight'})
        return synchronous_machines

    def _create_gen_characteristics_table(self, syn_gen_df_origin) -> pd.DataFrame:
        eq = self.cimConverter.cim['eq']
        if not eq['ReactiveCapabilityCurve'].empty and not eq['CurveData'].empty:
            if 'id_q_capability_curve_characteristic' not in syn_gen_df_origin.columns:
                    syn_gen_df_origin['id_q_capability_curve_characteristic'] = pd.Series(pd.NA, dtype="Int64")
            if 'q_capability_curve_table' not in self.cimConverter.net.keys():
                self.cimConverter.net['q_capability_curve_table'] = pd.DataFrame(
                    columns=['id_capability_curve', 'p_mw', 'q_min_mvar', 'q_max_mvar'])

            # get the curve data
            curve_data = self.cimConverter.cim['eq']['ReactiveCapabilityCurve'][['rdfId', 'curveStyle']].rename(
                columns={'rdfId': 'Curve', 'curveStyle': 'curve_style'})
            curve_points = pd.merge(curve_data, self.cimConverter.cim['eq']['CurveData'][
                ['rdfId', 'Curve', 'xvalue', 'y1value', 'y2value']], how='left', on='Curve')

            curve_points['id_q_capability_curve_characteristic'] = pd.factorize(curve_points['Curve'])[0]
            curve_points['id_q_capability_curve_characteristic'] = curve_points['id_q_capability_curve_characteristic'].astype('Int64')
            curve_points = curve_points.drop(columns=['rdfId']).rename(columns={'Curve': 'rdfId', 'xvalue': 'p_mw',
                                                                                'y1value': 'q_min_mvar',
                                                                                'y2value': 'q_max_mvar'})

            # Move 'id_q_capability_curve_characteristic' to the first column and save to net
            curve_points = curve_points[
                ['id_q_capability_curve_characteristic'] + [col for col in curve_points.columns if
                                                   col != 'id_q_capability_curve_characteristic']]
            self.cimConverter.net['q_capability_curve_table'] = curve_points
            self.cimConverter.net['q_capability_curve_table'] = (self.cimConverter.net['q_capability_curve_table'].drop
                                                                 (columns=['rdfId', 'curve_style']).rename(
                columns={'id_q_capability_curve_characteristic': 'id_q_capability_curve'}))

            # Drop unnecessary columns and duplicate rdfId
            curve_points = (
                curve_points.drop(columns=['p_mw', 'q_min_mvar', 'q_max_mvar']).drop_duplicates(subset=['rdfId']).
                rename(columns={'rdfId': 'InitialReactiveCapabilityCurve'}))

            syn_gen_df_origin = syn_gen_df_origin.drop(columns=['id_q_capability_curve_characteristic'])
            syn_gen_df_origin = pd.merge(syn_gen_df_origin, curve_points, how='left',
                                         on=['InitialReactiveCapabilityCurve'])

            # create reactive_capability_curve flag
            if 'reactive_capability_curve' not in syn_gen_df_origin.columns:
                syn_gen_df_origin['reactive_capability_curve'] = False
        return syn_gen_df_origin



from copy import deepcopy

from numpy import append, ceil

from pandapower.create import create_load, create_buses, create_line, create_empty_network, create_bus, \
    create_ext_grid, create_transformer
from pandapower.std_types import change_std_type, create_std_type
from pandapower.toolbox.element_selection import get_connected_elements
from pandapower.toolbox.power_factor import pq_from_cosphi


def _change_to_ohl(net, idx_busbar, new_lines, n_cable):
    """
    This function changes line types from cable to ohl beginning at the end of the feeders in a \
    way, that the tapped line has the most portions of overhead-lines.
    """
    con_lines = get_connected_elements(net, "line", idx_busbar) & new_lines
    cable_lines = con_lines
    last_con_lines = list(con_lines)

    while len(cable_lines) < n_cable:
        con_lines = sorted(get_connected_elements(
            net, "line", net.line.to_bus.loc[last_con_lines]) & new_lines - cable_lines)
        last_con_lines = deepcopy(con_lines)
        while len(con_lines) > 0:
            cable_lines.add(con_lines.pop(0))
    for idx_line in list(new_lines - cable_lines):
        change_std_type(net, idx_line, 'NFA2X 4x70', element="line")


def _create_loads_with_coincidence(net, buses):
    """
    This function creates loads to lv feeders with respect to coincidence factor c. c is \
    calculated by the number of feeder buses as described in Dickert, Schegner - \
    'Residential Load Models for Network Planning Purposes', Modern Electric Power Systems 2010, \
    Wroclaw, Poland.
    """
    # assumptions
    c_inf = 0.1
    P_max1 = 0.01  # MW
    powerfactor = 0.95  # in range of 0.9 to 1

    # calculations
    n_buses = len(buses)
    c = c_inf + (1 - c_inf) * n_buses ** (-1 / 2)
    p_mw = c * P_max1
    p_mw, q_mvar = pq_from_cosphi(p_mw, powerfactor, qmode='underexcited', pmode="load")

    # create loads
    for i in buses:
        create_load(net, i, p_mw=p_mw, q_mvar=q_mvar, sn_mva=P_max1)


def _create_feeder(net, net_data, branching, idx_busbar, linetype, lv_vn_kv):
    """
    This function creates the dickert lv network feeders by creating the buses and lines. In case \
    of branching the right position of branching must be found with respect to sum of feeder nodes \
    'n_DP' and the number of 'branching'.
    """
    n_DP = net_data[1]
    d_DP = net_data[0]
    buses = create_buses(net, int(n_DP), lv_vn_kv, zone='Feeder B' + str(branching),
                         type='m')
    from_bus = append(idx_busbar, buses[:-1])
    # branch consideration
    if branching == 1:
        n_LS = int(ceil(n_DP / 3))
        idx_B1 = 2 * n_LS
        if n_LS * 3 - n_DP == 2:
            idx_B1 -= 1
        from_bus[idx_B1] = buses[n_LS - 1]
    elif branching == 2:
        n_LS = int(ceil(n_DP / 6))
        idx_B1 = 3 * n_LS
        idx_B2 = 4 * n_LS
        if n_LS * 6 - n_DP >= 1:
            idx_B2 -= 1
        if n_LS * 6 - n_DP >= 4:
            idx_B1 -= 1
            idx_B2 -= 1
        if n_LS * 6 - n_DP == 5:
            idx_B2 -= 1
        from_bus[idx_B1] = buses[2 * n_LS - 1]
        from_bus[idx_B2] = buses[n_LS - 1]
    elif branching == 3:
        n_LS = int(ceil(n_DP / 10))
        idx_B1 = 4 * n_LS
        idx_B2 = 5 * n_LS
        idx_B3 = 7 * n_LS
        if n_LS * 10 - n_DP >= 1:
            idx_B2 -= 1
            idx_B3 -= 1
        if n_LS * 10 - n_DP >= 2:
            idx_B3 -= 1
        if n_LS * 10 - n_DP >= 3:
            idx_B3 -= 1
        if n_LS * 10 - n_DP >= 7:
            idx_B1 -= 1
            idx_B2 -= 1
            idx_B3 -= 1
        if n_LS * 10 - n_DP >= 8:
            idx_B2 -= 1
            idx_B3 -= 1
        if n_LS * 10 - n_DP == 9:
            idx_B3 -= 1
        from_bus[idx_B1] = buses[3 * n_LS - 1]
        from_bus[idx_B2] = buses[2 * n_LS - 1]
        from_bus[idx_B3] = buses[n_LS - 1]
    elif branching != 0:
        raise ValueError("branching must be in (0, 1, 2, 3), but is %s" % str(branching))

    # create lines
    new_lines = set()
    for i, f_bus in enumerate(from_bus):
        new_lines.add(create_line(net, f_bus, buses[i], length_km=d_DP * 1e-3,
                                  std_type='NAYY 4x150 SE'))

    # line type consideration
    if linetype == 'C&OHL':
        _change_to_ohl(net, idx_busbar, new_lines, round(len(new_lines) * 0.4))

    # create loads
    _create_loads_with_coincidence(net, buses)


def create_dickert_lv_feeders(net, busbar_index, feeders_range='short', linetype='cable',
                              customer='single', case='good'):
    """
    This function creates LV feeders from J. Dickert, M. Domagk and P. Schegner. "Benchmark \
    low voltage distribution networks based on cluster analysis of actual grid properties". \
    PowerTech, 2013 IEEE Grenoble.
    The number of this LV feeders will be in range of one to three, with respect to the optional \
    given parameters 'feeders_range', 'linetype', 'customer' and 'case'.
    The given 'preferred lines for feeders' are used, knowing that there are some other \
    standard types mentioned as well.

    Since the paper focusses on LV grids structure, load powers and MV connection are neglected, \
    so that the user should identify appropriate assumptions for trafo and load parameters.
    'trafo_type_name' and 'trafo_type_data' can be set directly by the user.
    By default, the load powers are calculated with coincidence factor, derived with normal \
    distributed peak system demand.

    INPUT:
        **net** (pandapowerNet) - The pandapower network to that the feeder will be connected to

        **busbar_index** (int) - The bus index of busbar, the feeders should be connected to

    OPTIONAL:

        **feeders_range** (str, 'short') - feeder length, which can be ('short', 'middle', 'long')

        **linetype** (str, 'cable') - the are different feeders provided for 'cable' or 'C&OHL'

        **customer** (str, 'single') - type of customers ('single' or 'multiple') supplied by the
            feeders

        **case** (str, 'good') - case of supply mission, which can be ('good', 'average', 'worse')

    EXAMPLE:
        >>> from pandapower.networks.dickert_lv_networks import create_dickert_lv_feeders, create_dickert_lv_networks
        >>> net = create_dickert_lv_network()
        >>> create_dickert_lv_feeders(net, busbar_index=1, customer='multiple')
    """
    # --- paper data - TABLE III and IV
    parameters = {'short': {'cable': {'single': {'good': [60, 1, False, False, False],
                                                 'average': [120, 1, False, False, False],
                                                 'worse': [80, 2, False, False, False]},
                                      'multiple': {'good': [80, 3, True, False, False],
                                                   'average': [50, 6, True, False, False],
                                                   'worse': [40, 10, True, False, False]}}},
                  'middle': {'cable': {'multiple': {'good': [40, 15, True, True, False],
                                                    'average': [35, 20, True, True, False],
                                                    'worse': [30, 25, True, True, False]}},
                             'C&OHL': {'multiple': {'good': [50, 10, True, True, False],
                                                    'average': [45, 13, True, True, False],
                                                    'worse': [40, 16, True, True, False]}}},
                  'long': {'cable': {'multiple': {'good': [30, 30, False, True, True],
                                                  'average': [30, 40, False, True, True],
                                                  'worse': [30, 50, False, True, True]}},
                           'C&OHL': {'multiple': {'good': [40, 20, False, True, True],
                                                  'average': [40, 30, False, True, True],
                                                  'worse': [40, 40, False, True, True]}}}}
    # process network choosing input data
    try:
        case = case if case != "bad" else "worse"
        net_data = parameters[feeders_range][linetype][customer][case]
    except KeyError:
        raise ValueError("This combination of 'feeders_range', 'linetype', 'customer' and 'case' "
                         "is no dickert network.")

    # add missing line types
    if 'NFA2X 4x70' not in net.std_types['line'].keys():
        create_std_type(net, {"c_nf_per_km": 12.8, "r_ohm_per_km": 0.443, "x_ohm_per_km": 0.07,
                              "max_i_ka": 0.205, "type": "ol"}, name='NFA2X 4x70',
                        element="line")
    # determine low voltage vn_kv
    lv_vn_kv = net.bus.vn_kv.at[busbar_index]

    # feeder without branch line
    _create_feeder(net, net_data, 0, busbar_index, linetype, lv_vn_kv)
    # feeder with one branch line
    if net_data[2]:
        _create_feeder(net, net_data, 1, busbar_index, linetype, lv_vn_kv)
    # feeder with two branch lines
    if net_data[3]:
        _create_feeder(net, net_data, 2, busbar_index, linetype, lv_vn_kv)
    # feeder with three branch lines
    if net_data[4]:
        _create_feeder(net, net_data, 3, busbar_index, linetype, lv_vn_kv)


def create_dickert_lv_network(feeders_range='short', linetype='cable', customer='single',
                              case='good', trafo_type_name='0.4 MVA 20/0.4 kV',
                              trafo_type_data=None):
    """
    This function creates a LV network from J. Dickert, M. Domagk and P. Schegner. "Benchmark \
    low voltage distribution networks based on cluster analysis of actual grid properties". \
    PowerTech, 2013 IEEE Grenoble.
    This LV network will have one to three feeders connected to MV-LV-Trafo. To connect more feeders
    with respect to the optional given parameters 'feeders_range', 'linetype', 'customer' and
    'case', the 'create_dickert_lv_feeders' function can be executed.
    The given 'preferred lines for feeders' are used, knowing that there are some other \
    standard types mentioned as well.

    Since the paper focusses on LV grids structure, load powers and MV connection are neglected, \
    so that the user should identify appropriate assumptions for trafo and load parameters.
    'trafo_type_name' and 'trafo_type_data' can be set directly by the user.
    By default, the load powers are calculated with coincidence factor, derived with normal \
    distributed peak system demand, described in Dickert, Schegner - \
    'Residential Load Models for Network Planning Purposes', Modern Electric Power Systems 2010, \
    Wroclaw, Poland, with the given example assumptions:

    - c_inf = 0.1
    - P_max1 = 10 kW
    - powerfactor = 0.95 ind. (in range of 0.9 to 1)

    OPTIONAL:

        **feeders_range** (str, 'short') - feeder length, which can be ('short', 'middle', 'long')

        **linetype** (str, 'cable') - the are different feeders provided for 'cable' or 'C&OHL'

        **customer** (str, 'single') - type of customers ('single' or 'multiple') supplied by the \
            feeders

        **case** (str, 'good') - case of supply mission, which can be ('good', 'average', 'bad')

        **trafo_type_name** (str, '0.4 MVA 20/0.4 kV') - name of the HV-MV-Trafo standard type

        **trafo_type_data** (dict, None) - if 'trafo_type_name' is not in pandapower standard \
            types, the data of this new trafo types must be given here in pandapower trafo type way

    OUTPUT:

        **net** (pandapowerNet) - Returns the required dickert lv network

    EXAMPLE:
        >>> from pandapower.networks.dickert_lv_networks import create_dickert_lv_network
        >>> net = create_dickert_lv_network()
    """
    # --- create network
    net = create_empty_network(name='dickert_lv_network with' + feeders_range +
                                    '-range feeders, ' + linetype + 'and ' + customer +
                                    'customers in ' + case + 'case')
    # assumptions
    mv_vn_kv = 20
    lv_vn_kv = 0.4

    # create mv connection
    mv_bus = create_bus(net, mv_vn_kv, name='mv bus')
    busbar_index = create_bus(net, lv_vn_kv, name='busbar')
    create_ext_grid(net, mv_bus)
    if trafo_type_name not in net.std_types['trafo'].keys():
        create_std_type(net, trafo_type_data, name=trafo_type_name, element="trafo")
    create_transformer(net, mv_bus, busbar_index, std_type=trafo_type_name)

    # create feeders
    create_dickert_lv_feeders(net=net, busbar_index=busbar_index, feeders_range=feeders_range,
                              linetype=linetype, customer=customer, case=case)

    return net


if __name__ == "__main__":
    if 0:
        feeders_range = 'middle'
        linetype = 'C&OHL'
        customer = 'multiple'
        case = 'bad'
        trafo_type_name = '0.4 MVA 20/0.4 kV'
        trafo_type_data = None
        net = create_dickert_lv_network(feeders_range=feeders_range, linetype=linetype,
                                        customer=customer, case=case,
                                        trafo_type_name=trafo_type_name,
                                        trafo_type_data=trafo_type_data)
        from pandapower.plotting import simple_plot

        simple_plot(net)
    else:
        pass



import logging
import time

import pandas as pd
import numpy as np

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.nonLinearShuntCompensatorCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class NonLinearShuntCompensatorCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_nonlinear_shunt_compensator_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting NonlinearShuntCompensator.")
        if self.cimConverter.cim['eq']['NonlinearShuntCompensator'].index.size > 0:
            eqssh_shunts = self._prepare_nonlinear_shunt_compensator_cim16()
            self._create_shunt_characteristic_table(eqssh_shunts)
            self.cimConverter.copy_to_pp('shunt', eqssh_shunts)
        else:
            eqssh_shunts = pd.DataFrame(None)
        self.logger.info("Created %s shunts in %ss." % (eqssh_shunts.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s shunts from NonlinearShuntCompensator in %ss." %
                    (eqssh_shunts.index.size, time.time() - time_start)))

    def _prepare_nonlinear_shunt_compensator_cim16(self) -> pd.DataFrame:
        eqssh_shunts = self.cimConverter.merge_eq_ssh_profile('NonlinearShuntCompensator', add_cim_type_column=True)
        eqssh_shunts = pd.merge(eqssh_shunts, self.cimConverter.bus_merge, how='left', on='rdfId')

        eqssh_shunts['p'] = float('NaN')
        eqssh_shunts['q'] = float('NaN')
        eqssh_shunts_cols = eqssh_shunts.columns.to_list()
        nscp = self.cimConverter.cim['eq']['NonlinearShuntCompensatorPoint'][
            ['NonlinearShuntCompensator', 'sectionNumber', 'b', 'g']].rename(
            columns={'NonlinearShuntCompensator': 'rdfId'})
        # calculate p and q from b, g, and all the sections
        for i in range(1, int(nscp['sectionNumber'].max()) + 1):
            nscp_t = nscp.loc[nscp['sectionNumber'] == i]
            eqssh_shunts = pd.merge(eqssh_shunts, nscp_t, how='left', on='rdfId')
            y = eqssh_shunts['g'] + eqssh_shunts['b'] * 1j
            s = eqssh_shunts['nomU'] ** 2 * np.conj(y)
            eqssh_shunts['p_temp'] = s.values.real
            eqssh_shunts['q_temp'] = s.values.imag
            if i == 1:
                eqssh_shunts['p'] = eqssh_shunts['p_temp'][:]
                eqssh_shunts['q'] = eqssh_shunts['q_temp'][:]
            else:
                eqssh_shunts.loc[eqssh_shunts['sections'] >= eqssh_shunts['sectionNumber'], 'p'] = \
                    eqssh_shunts['p'] + eqssh_shunts['p_temp']
                eqssh_shunts.loc[eqssh_shunts['sections'] >= eqssh_shunts['sectionNumber'], 'q'] = \
                    eqssh_shunts['q'] + eqssh_shunts['q_temp']
            eqssh_shunts = eqssh_shunts[eqssh_shunts_cols]
        if 'inService' in eqssh_shunts.columns:
            eqssh_shunts['connected'] = eqssh_shunts['connected'] & eqssh_shunts['inService']
        # added to use the logic of p/q values multiplied by the number of steps
        # this is only correct for the current step values
        eqssh_shunts['p'] = eqssh_shunts['p'] / eqssh_shunts['sections']
        eqssh_shunts['q'] = eqssh_shunts['q'] / eqssh_shunts['sections']
        eqssh_shunts = eqssh_shunts.rename(columns={
            'rdfId': sc['o_id'], 'rdfId_Terminal': sc['t'], 'connected': 'in_service', 'index_bus': 'bus',
            'nomU': 'vn_kv', 'p': 'p_mw', 'q': 'q_mvar', 'sections': 'step', 'maximumSections': 'max_step'})
        return eqssh_shunts

    def _create_shunt_characteristic_table(self, eqssh_shunts):
        if 'id_characteristic_table' not in eqssh_shunts.columns:
            eqssh_shunts['id_characteristic_table'] = np.nan
        if 'shunt_characteristic_table' not in self.cimConverter.net.keys():
            self.cimConverter.net['shunt_characteristic_table'] = pd.DataFrame(
                columns=['id_characteristic', 'step', 'q_mvar', 'p_mw'])
        char_temp = eqssh_shunts.drop(columns=['p_mw', 'q_mvar', 'step'])
        char_temp['p'] = float('NaN')
        char_temp['q'] = float('NaN')
        # get the NonlinearShuntCompensatorPoints
        nscp = self.cimConverter.cim['eq']['NonlinearShuntCompensatorPoint'][
            ['NonlinearShuntCompensator', 'sectionNumber', 'b', 'g']].rename(
            columns={'NonlinearShuntCompensator': sc['o_id']})
        char_temp = pd.merge(char_temp, nscp, how='left', on=sc['o_id'])
        # calculate p & q from b & g for all sections
        y = char_temp['g'] + char_temp['b'] * 1j
        s = char_temp['vn_kv'] ** 2 * np.conj(y)
        char_temp['p_temp'] = s.values.real
        char_temp['q_temp'] = s.values.imag
        # calculate cumulative sums for all sections
        char_temp = char_temp.sort_values(by=[sc['o_id'], 'sectionNumber'])
        char_temp['p'] = char_temp.groupby(sc['o_id'])['p_temp'].cumsum()
        char_temp['q'] = char_temp.groupby(sc['o_id'])['q_temp'].cumsum()
        char_temp = char_temp.rename(columns={'p': 'p_mw', 'q': 'q_mvar', 'sectionNumber': 'step'})
        char_temp['step'] = char_temp['step'].astype(int)
        # assign id_characteristic
        char_temp['id_characteristic'] = pd.factorize(char_temp[sc['o_id']])[0]

        # set the id_characteristic at the corresponding shunt
        id_char_dict = char_temp.drop_duplicates(sc['o_id']).set_index(sc['o_id'])['id_characteristic'].to_dict()
        eqssh_shunts['id_characteristic_table'] = eqssh_shunts[sc['o_id']].map(id_char_dict).astype('Int64')

        # create step_dependency_table flag
        if 'step_dependency_table' not in eqssh_shunts.columns:
            eqssh_shunts["step_dependency_table"] = False
        # set step_dependency_table as True for all non-linear shunt compensators
        eqssh_shunts.loc[eqssh_shunts['id_characteristic_table'].notna(), 'step_dependency_table'] = True

        # populate shunt_characteristic_temp table
        self.cimConverter.net['shunt_characteristic_table'] = \
            char_temp[['id_characteristic', 'step', 'q_mvar', 'p_mw']]



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import random as rd

from pandapower.create import create_empty_network, create_bus, create_ext_grid, create_line, create_load, \
    create_transformer
from pandapower.std_types import create_std_type


# --- support functions

def _create_empty_network_with_transformer(trafotype, V_OS=10., V_US=0.4):
    """
    Creates a Network with transformer and infeeder. The reference bus on the \
    high-voltage side is called "Trafostation_OS". The bus on the low-voltage \
    side is called "main_busbar".
    The voltage levels can be set manually and the transformer parameter can \
    be set with "ti"
    """
    pd_net = create_empty_network()
    NFA2X4x70 = {"c_nf_per_km": 0, "r_ohm_per_km": 0.443, "x_ohm_per_km": 0.069, "max_i_ka": 0.270,
                 "type": "ol", "q_mm2": 70}
    NAYY4x50 = {"c_nf_per_km": 670, "r_ohm_per_km": 0.6417, "x_ohm_per_km": 0.084823,
                "max_i_ka": 0.141, "type": "cs", "q_mm2": 50}
    NAYY4x150 = {"c_nf_per_km": 830, "r_ohm_per_km": 0.2067, "x_ohm_per_km": 0.08042478,
                 "max_i_ka": 0.275, "type": "cs", "q_mm2": 150}
    NAYY4x185 = {"c_nf_per_km": 830, "r_ohm_per_km": 0.165, "x_ohm_per_km": 0.08042478,
                 "max_i_ka": 0.313, "type": "cs", "q_mm2": 185}
    NYY4x35 = {"c_nf_per_km": 0, "r_ohm_per_km": 0.5240284, "x_ohm_per_km": 0.08513716,
               "max_i_ka": 0.156, "type": "cs", "q_mm2": 35}
    create_std_type(net=pd_net, data=NFA2X4x70, name="NFA2X 4x70", element="line")
    create_std_type(net=pd_net, data=NAYY4x50, name="NAYY 4x50", element="line")
    create_std_type(net=pd_net, data=NAYY4x150, name="NAYY 4x150", element="line")
    create_std_type(net=pd_net, data=NAYY4x185, name="NAYY 4x185", element="line")
    create_std_type(net=pd_net, data=NYY4x35, name="NYY 4x35", element="line")
    T100kVA = {"sn_mva": 0.100, "vn_hv_kv": 10, "vn_lv_kv": 0.4, "vk_percent": 4,
               "vkr_percent": 1.2, "pfe_kw": 0.45, "i0_percent": 0.25, "shift_degree": 150,
               "vector_group": "Dyn5"}
    T160kVA = {"sn_mva": 0.160, "vn_hv_kv": 10, "vn_lv_kv": 0.4, "vk_percent": 4,
               "vkr_percent": 1.2, "pfe_kw": 0.38, "i0_percent": 0.26, "shift_degree": 150,
               "vector_group": "Dyn5"}
    create_std_type(net=pd_net, data=T100kVA, name="0.1 MVA 10/0.4 kV", element="trafo")
    create_std_type(net=pd_net, data=T160kVA, name="0.16 MVA 10/0.4 kV", element="trafo")

    busnr1 = create_bus(pd_net, name="Trafostation_OS", vn_kv=V_OS)
    create_ext_grid(pd_net, bus=busnr1)
    main_busbar_nr = create_bus(pd_net, name="main_busbar", vn_kv=V_US, type="b")
    create_transformer(pd_net, hv_bus=busnr1, lv_bus=main_busbar_nr, std_type=trafotype,
                          name="trafo 1")
    return pd_net, main_busbar_nr


def _add_lines_and_loads(pd_net, n_lines, startbusnr, length_per_line,
                         std_type="NAYY 4x150 SE", p_load_mw=0,
                         q_load_mvar=0, branchnr=1,
                         l_para_per_km=None):
    """
    Creates a single unsplitted branch on the startbus of n lines. It \
    sequencely adds lines, buses and loads.

    Loads will only be added if p_load_mw or q_load_mvar \
    is assigned

    The branch number could be assigned with branchnr. It will be added to \
    the name ti keep track on the node position
    """

    startpoint_bus = 1
    startpoint_line = 1
    bus_before = startbusnr
    for i in list(range(n_lines)):
        buscounter = startpoint_bus + i
        linecounter = startpoint_line + i
        created_bus_nr = create_bus(pd_net, name="bus_%d_%d" % (branchnr, buscounter), vn_kv=.4)

        create_line(pd_net, bus_before, created_bus_nr, length_km=length_per_line,
                       name="line_%d_%d" % (branchnr, linecounter), std_type=std_type)

        if p_load_mw or q_load_mvar:
            create_load(pd_net, created_bus_nr, p_mw=p_load_mw, q_mvar=q_load_mvar)

        bus_before = created_bus_nr  # rueckgefuehrter Wert in der Schleife

    return pd_net


def _add_lines_with_branched_loads(net, n_lines, startbus, length_per_line,
                                   std_type="NAYY 4x150 SE", p_load_mw=0,
                                   q_load_mvar=0,
                                   length_branchout_line_1=0.022,
                                   length_branchout_line_2=0,
                                   std_type_branchout_line_1="NAYY 4x50 SE",
                                   std_type_branchout_line_2="NAYY 4x50 SE",
                                   prob_branchout_line_1=0.5, branchnr=1):
    """
    Creates a single unsplitted branch on the startbus. each bus on the main \
    line is connected to a branch out line which connects \
    the loadbus (households).

    If there are two std_types given for the branch_out_lin. The cable_types \
    interchange with the given probability

    If there are two lengths of branchoutlines are given, the \
    lengths interchange.
    It begins with length 1 and switches to length 2. The cable with length 1 \
    is named as "MUF_" and length 2 becomes "KV_".

    Loads will only be added if p_load_mw or q_load_mvar \
    is assigned

    The branch number could be assigned with branchnr. It will be added to the\
     name ti keep track on the node position
    """

    # support function
    startpoint_bus = 1
    startpoint_line = 1
    bus_before = startbus
    length_branchout_line = length_branchout_line_1
    # destinct between Muffe und Kabelverteiler
    if length_branchout_line_2:
        bustype = "MUF"
    else:
        bustype = "bus"
    std_type_branchout_line = std_type_branchout_line_1
    for i in range(n_lines):
        buscounter = startpoint_bus + i
        linecounter = startpoint_line + i
        created_bus_nr = create_bus(net, name="%s_%d_%d" % (bustype, branchnr, buscounter),
                                       type="b" if bustype == "KV" else "n", vn_kv=.4)
        create_line(net, bus_before, created_bus_nr,
                       length_km=length_per_line,
                       name="line_%d_%d" % (branchnr, linecounter),
                       std_type=std_type)

        loadbusnr = create_bus(net, name="loadbus_%d_%d" % (branchnr, buscounter), vn_kv=.4)

        create_line(net, created_bus_nr, loadbusnr,
                       length_km=length_branchout_line,
                       name="branchout_line_%d_%d" % (branchnr, linecounter),
                       std_type=std_type_branchout_line)

        if p_load_mw or q_load_mvar:
            create_load(net, loadbusnr,
                           p_mw=p_load_mw, q_mvar=q_load_mvar)

        bus_before = created_bus_nr  # rueckgefuehrter Wert in der Schleife

        # alternates the length of the branch out lines if needed
        if length_branchout_line_2:
            if length_branchout_line == length_branchout_line_1:
                length_branchout_line = length_branchout_line_2
                bustype = "KV"
            else:
                length_branchout_line = length_branchout_line_1
                bustype = "MUF"
        #  changes branch out lines according to the probabillity if needed
        if std_type_branchout_line_2:
            if rd.random() > prob_branchout_line_1:
                std_type_branchout_line = std_type_branchout_line_2
            else:
                std_type_branchout_line = std_type_branchout_line_1
    return net

def _create_branch_network(trafotype, v_os, num_lines, len_lines, **kwargs):
    pd_net, main_busbar_nr = _create_empty_network_with_transformer(trafotype, V_OS=v_os)
    for i, (n_line, len_line) in enumerate(zip(num_lines, len_lines), start=1):
        _add_lines_and_loads(pd_net, n_line, startbusnr=main_busbar_nr,
                             length_per_line=len_line, branchnr=i, **kwargs)
    return pd_net


def _create_branched_loads_network(trafotype, v_os, num_lines, len_lines, **kwargs):
    pd_net, main_busbar_nr = _create_empty_network_with_transformer(trafotype, V_OS=v_os)
    for i, (n_line, len_line) in enumerate(zip(num_lines, len_lines), start=1):
        _add_lines_with_branched_loads(pd_net, n_line, startbus=main_busbar_nr,
                                       length_per_line=len_line, branchnr=i, **kwargs)
    return pd_net

# --- main functions

def create_kerber_landnetz_freileitung_1(n_lines=13,
                                         l_lines_in_km=0.021, std_type="NFA2X 4x70",
                                         trafotype="0.16 MVA 10/0.4 kV",
                                         p_load_mw=0.008, q_load_mvar=0, v_os=10.):

    pd_net, main_busbar_nr = _create_empty_network_with_transformer(trafotype, V_OS=v_os)
    _add_lines_and_loads(pd_net, n_lines, startbusnr=main_busbar_nr,
                         length_per_line=l_lines_in_km, std_type=std_type,
                         p_load_mw=p_load_mw,
                         q_load_mvar=q_load_mvar)
    return pd_net


def create_kerber_landnetz_freileitung_2(n_branch_1=6, n_branch_2=2,
                                         l_lines_1_in_km=0.038,
                                         l_lines_2_in_km=0.081,
                                         std_type="NFA2X 4x70",
                                         trafotype="0.1 MVA 10/0.4 kV",
                                         p_load_mw=0.008, q_load_mvar=0,
                                         v_os=10.):
    num_lines = [n_branch_1, n_branch_2]
    len_lines = [l_lines_1_in_km, l_lines_2_in_km]
    pd_net = _create_branch_network(trafotype=trafotype, v_os=v_os, num_lines=num_lines,
                                    len_lines=len_lines, std_type=std_type,
                                    p_load_mw=p_load_mw, q_load_mvar=q_load_mvar)
    return pd_net


def create_kerber_landnetz_kabel_1(n_branch_1=6, n_branch_2=2, l_lines_1_in_km=0.082,
                                   l_lines_2_in_km=0.175, std_type="NAYY 4x150",
                                   std_type_branchout_line="NAYY 4x50",
                                   trafotype="0.1 MVA 10/0.4 kV", p_load_mw=0.008,
                                   q_load_mvar=0., length_branchout_line_1=0.018,
                                   length_branchout_line_2=0.033, v_os=10.):
    """
    .. note:: It is assumed that every second bus in a branch is a "KV".
    """
    num_lines = [n_branch_1, n_branch_2]
    len_lines = [l_lines_1_in_km, l_lines_2_in_km]
    pd_net = _create_branched_loads_network(trafotype=trafotype, v_os=v_os, num_lines=num_lines,
                                            len_lines=len_lines, std_type=std_type,
                                            p_load_mw=p_load_mw,
                                            q_load_mvar=q_load_mvar,
                                            length_branchout_line_1=length_branchout_line_1,
                                            length_branchout_line_2=length_branchout_line_2,
                                            std_type_branchout_line_1=std_type_branchout_line)
    return pd_net


def create_kerber_landnetz_kabel_2(n_branch_1=12, n_branch_2=2, l_lines_1_in_km=0.053,
                                   l_lines_2_in_km=0.175, std_type="NAYY 4x150",
                                   trafotype="0.16 MVA 10/0.4 kV", p_load_mw=0.008,
                                   q_load_mvar=0., length_branchout_line_1=0.018,
                                   length_branchout_line_2=0.033,
                                   std_type_branchout_line="NAYY 4x50", v_os=10.):
    """
    .. note:: It is assumed that every second bus in a branch is a "KV".
    """
    num_lines = [n_branch_1, n_branch_2]
    len_lines = [l_lines_1_in_km, l_lines_2_in_km]
    pd_net = _create_branched_loads_network(trafotype=trafotype, v_os=v_os, num_lines=num_lines,
                                            len_lines=len_lines, std_type=std_type,
                                            p_load_mw=p_load_mw,
                                            q_load_mvar=q_load_mvar,
                                            length_branchout_line_1=length_branchout_line_1,
                                            length_branchout_line_2=length_branchout_line_2,
                                            std_type_branchout_line_1=std_type_branchout_line)
    return pd_net


def create_kerber_dorfnetz(std_type="NAYY 4x150", trafotype="0.4 MVA 10/0.4 kV",
                           p_load_mw=0.006,
                           q_load_mvar=0., length_branchout_line_1=0.015,
                           length_branchout_line_2=0.031,
                           std_type_branchout_line="NAYY 4x50", v_os=10.):
    """
    .. note:: It is assumed that every second bus in a branch is a "KV".
    """
    num_lines = [9, 9, 16, 12, 7, 4]
    len_lines = [0.04, 0.04, 0.029, 0.032, 0.043, 0.064]
    pd_net = _create_branched_loads_network(trafotype=trafotype, v_os=v_os, num_lines=num_lines,
                                            len_lines=len_lines, std_type=std_type,
                                            p_load_mw=p_load_mw,
                                            q_load_mvar=q_load_mvar,
                                            length_branchout_line_1=length_branchout_line_1,
                                            length_branchout_line_2=length_branchout_line_2,
                                            std_type_branchout_line_1=std_type_branchout_line)
    return pd_net


def create_kerber_vorstadtnetz_kabel_1(std_type="NAYY 4x150", p_load_mw=0.002, q_load_mvar=0.,
                                       trafotype="0.63 MVA 10/0.4 kV", v_os=10.):
    """
    .. note:: Please pay attention, that the linetypes of the branch out house connections are \
    randomly distributed according to the probability 50% between "NAYY 4x50" and "NYY 4x35"
    """
    num_lines = [14, 14, 14, 19, 19, 10, 10, 10, 32, 4]
    len_lines = [0.021, 0.021, 0.021, 0.017, 0.017, 0.025, 0.025, 0.025, 0.011, 0.060]
    pd_net = _create_branched_loads_network(trafotype=trafotype, v_os=v_os, num_lines=num_lines,
                                            len_lines=len_lines, std_type=std_type,
                                            p_load_mw=p_load_mw,
                                            q_load_mvar=q_load_mvar,
                                            length_branchout_line_1=0.011,
                                            std_type_branchout_line_1="NAYY 4x50",
                                            std_type_branchout_line_2="NYY 4x35",
                                            prob_branchout_line_1=0.5)
    return pd_net


def create_kerber_vorstadtnetz_kabel_2(std_type="NAYY 4x185", p_load_mw=.002, q_load_mvar=0.,
                                       trafotype="0.63 MVA 10/0.4 kV", v_os=10.):
    """
    .. note:: Please pay attention, that the linetypes of the branch out house \
    connections are randomlydistributed according to the probability 50% between \
    "NAYY 50" and "NYY 35"
    """

    num_lines = [15, 15, 15, 20, 20, 11, 11, 32, 5]
    len_lines = [0.023, 0.023, 0.023, 0.020, 0.020, 0.026, 0.026, 0.014, 0.050]
    pd_net = _create_branched_loads_network(trafotype=trafotype, v_os=v_os, num_lines=num_lines,
                                            len_lines=len_lines, std_type=std_type,
                                            p_load_mw=p_load_mw,
                                            q_load_mvar=q_load_mvar,
                                            length_branchout_line_1=0.011,
                                            std_type_branchout_line_1="NAYY 4x50",
                                            std_type_branchout_line_2="NYY 4x35",
                                            prob_branchout_line_1=0.5)
    return pd_net





# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from numbers import Number

import numpy as np
from scipy.sparse.linalg import factorized

from pandapower.auxiliary import _clean_up, _add_ppc_options, _add_sc_options, _add_auxiliary_elements
from pandapower.pd2ppc_zero import _pd2ppc_zero
from pandapower.pypower.idx_brch_sc import K_ST
from pandapower.results import _copy_results_ppci_to_ppc
from pandapower.results import init_results
from pandapower.shortcircuit.currents import _calc_ikss, \
    _calc_ikss_1ph, _calc_ip, _calc_ith, _calc_branch_currents, _calc_branch_currents_complex
from pandapower.shortcircuit.impedance import _calc_zbus, _calc_ybus, _calc_rx
from pandapower.shortcircuit.kappa import _add_kappa_to_ppc
from pandapower.shortcircuit.ppc_conversion import _init_ppc, _create_k_updated_ppci, _get_is_ppci_bus
from pandapower.shortcircuit.results import _extract_results, _copy_result_to_ppci_orig

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def calc_sc(net, bus=None,
            fault="3ph", case='max', lv_tol_percent=10, topology="auto", ip=False,
            ith=False, tk_s=1., kappa_method="C", r_fault_ohm=0., x_fault_ohm=0.,
            branch_results=False, check_connectivity=True, return_all_currents=False,
            inverse_y=True, use_pre_fault_voltage=False):
    """
    Calculates minimal or maximal symmetrical short-circuit currents.
    The calculation is based on the method of the equivalent voltage source
    according to DIN/IEC EN 60909.
    The initial short-circuit alternating current *ikss* is the basis of the short-circuit
    calculation and is therefore always calculated.
    Other short-circuit currents can be calculated from *ikss* with the conversion factors defined
    in DIN/IEC EN 60909.

    The output is stored in the net.res_bus_sc table as a short_circuit current
    for each bus.

    INPUT:
        **net** (pandapowerNet) pandapower Network

        **bus** (int, list, np.array, None) defines if short-circuit calculations should only be calculated for
        defined bus

        ***fault** (str, 3ph) type of fault

            - "3ph" for three-phase

            - "2ph" for two-phase (phase-to-phase) short-circuits

            - "1ph" for single-phase-to-ground faults

        **case** (str, "max")

            - "max" for maximal current calculation

            - "min" for minimal current calculation

        **lv_tol_percent** (int, 10) voltage tolerance in low voltage grids

            - 6 for 6% voltage tolerance

            - 10 for 10% voltage olerance

        **ip** (bool, False) if True, calculate aperiodic short-circuit current

        **ith** (bool, False) if True, calculate equivalent thermical short-circuit current Ith

        **topology** (str, "auto") define option for meshing (only relevant for ip and ith)

            - "meshed" - it is assumed all buses are supplied over multiple paths

            - "radial" - it is assumed all buses are supplied over exactly one path

            - "auto" - topology check for each bus is performed to see if it is supplied over multiple paths

        **tk_s** (float, 1) failure clearing time in seconds (only relevant for ith)

        **r_fault_ohm** (float, 0) fault resistance in Ohm

        **x_fault_ohm** (float, 0) fault reactance in Ohm

        **branch_results** (bool, False) defines if short-circuit results should also be generated for branches

        **return_all_currents** (bool, False) applies only if branch_results=True, if True short-circuit currents for
        each (branch, bus) tuple is returned otherwise only the max/min is returned

        **inverse_y** (bool, True) defines if complete inverse should be used instead of LU factorization,
        factorization version is in experiment which should be faster and memory efficienter

        **use_pre_fault_voltage** (bool, False) whether to consider the pre-fault grid state (superposition method,
        "Type C")


    OUTPUT:

    EXAMPLE:
        calc_sc(net)

        print(net.res_bus_sc)
    """
    if fault not in ["3ph", "2ph", "1ph"]:
        raise NotImplementedError(
            "Only 3ph, 2ph and 1ph short-circuit currents implemented")

    if len(net.gen) and (ip or ith):
        logger.warning("aperiodic, thermal short-circuit currents are only implemented for "
                       "faults far from generators!")

    if case not in ['max', 'min']:
        raise ValueError('case can only be "min" or "max" for minimal or maximal short "\
                                "circuit current')

    if topology not in ["meshed", "radial", "auto"]:
        raise ValueError(
            'specify network structure as "meshed", "radial" or "auto"')

    if branch_results:
        logger.warning("Branch results are in beta mode and might not always be reliable, "
                       "especially for transformers")

    if use_pre_fault_voltage:
        init_vm_pu = init_va_degree = "results"
        trafo_model = net._options["trafo_model"]  # trafo model for SC must match the trafo model for PF calculation
        if not isinstance(bus, Number) and len(net.sgen.query("in_service")) > 0:
            raise NotImplementedError("Short-circuit with Type C method and sgen is only implemented for a single bus")
    else:
        init_vm_pu = init_va_degree = "flat"
        trafo_model = "pi"

    # Convert bus to numpy array
    if bus is None:
        bus = net.bus.index.values
    else:
        bus = np.array([bus]).ravel()

    kappa = ith or ip
    net["_options"] = {}
    _add_ppc_options(net, calculate_voltage_angles=False, trafo_model=trafo_model,
                     check_connectivity=check_connectivity, mode="sc", switch_rx_ratio=2,
                     init_vm_pu=init_vm_pu, init_va_degree=init_va_degree, enforce_q_lims=False,
                     recycle=None)
    _add_sc_options(net, fault=fault, case=case, lv_tol_percent=lv_tol_percent, tk_s=tk_s, topology=topology,
                    r_fault_ohm=r_fault_ohm, x_fault_ohm=x_fault_ohm, kappa=kappa, ip=ip, ith=ith,
                    branch_results=branch_results, kappa_method=kappa_method, return_all_currents=return_all_currents,
                    inverse_y=inverse_y, use_pre_fault_voltage=use_pre_fault_voltage)
    init_results(net, "sc")

    if fault in ("2ph", "3ph"):
        _calc_sc(net, bus)
    elif fault == "1ph":
        _calc_sc_1ph(net, bus)
    else:
        raise ValueError("Invalid fault %s" % fault)


def _calc_current(net, ppci_orig, bus):
    # Select required ppci bus
    ppci_bus = _get_is_ppci_bus(net, bus)

    # update ppci
    non_ps_gen_ppci_bus, non_ps_gen_ppci, ps_gen_bus_ppci_dict = \
        _create_k_updated_ppci(net, ppci_orig, ppci_bus=ppci_bus)

    # For each ps_gen_bus one unique ppci is required
    ps_gen_ppci_bus = list(ps_gen_bus_ppci_dict.keys())

    for calc_bus in ps_gen_ppci_bus + [non_ps_gen_ppci_bus]:
        if isinstance(calc_bus, np.ndarray):
            # Use ppci for general bus
            this_ppci, this_ppci_bus = non_ps_gen_ppci, calc_bus
        else:
            # Use specific ps_gen_bus ppci
            this_ppci, this_ppci_bus = ps_gen_bus_ppci_dict[calc_bus], np.array([calc_bus])

        _calc_ybus(this_ppci)
        if net["_options"]["inverse_y"]:
            _calc_zbus(net, this_ppci)
        else:
            # Factorization Ybus once
            # scipy.sparse.linalg.factorized converts the input matrix to csc from csr and raises a warning
            # todo: create Ybus in CSC format instead of CSR format if known that inverse_y is False?
            this_ppci["internal"]["ybus_fact"] = factorized(this_ppci["internal"]["Ybus"].tocsc())

        _calc_rx(net, this_ppci, this_ppci_bus)
        _calc_ikss(net, this_ppci, this_ppci_bus)
        _add_kappa_to_ppc(net, this_ppci)
        if net["_options"]["ip"]:
            _calc_ip(net, this_ppci)
        if net["_options"]["ith"]:
            _calc_ith(net, this_ppci)

        if net._options["branch_results"]:
            if net._options["fault"] == "3ph":
                _calc_branch_currents_complex(net, this_ppci, this_ppci_bus)
            else:
                _calc_branch_currents(net, this_ppci, this_ppci_bus)

        _copy_result_to_ppci_orig(ppci_orig, this_ppci, this_ppci_bus,
                                  calc_options=net._options)


def _calc_sc(net, bus):
    ppc, ppci = _init_ppc(net)

    _calc_current(net, ppci, bus)

    ppc = _copy_results_ppci_to_ppc(ppci, ppc, "sc")
    _extract_results(net, ppc, ppc_0=None, bus=bus)
    _clean_up(net)

    if "ybus_fact" in ppci["internal"]:
        # Delete factorization object
        ppci["internal"].pop("ybus_fact")


def _calc_sc_1ph(net, bus):
    """
    calculation method for single phase to ground short-circuit currents
    """
    _add_auxiliary_elements(net)
    # pos. seq bus impedance
    ppc, ppci = _init_ppc(net)
    # Create k updated ppci
    ppci_bus = _get_is_ppci_bus(net, bus)
    _, ppci, _ = _create_k_updated_ppci(net, ppci, ppci_bus=ppci_bus)
    _calc_ybus(ppci)

    # zero seq bus impedance
    ppc_0, ppci_0 = _pd2ppc_zero(net, ppc['branch'][:, K_ST])
    _calc_ybus(ppci_0)

    if net["_options"]["inverse_y"]:
        _calc_zbus(net, ppci)
        _calc_zbus(net, ppci_0)
    else:
        # Factorization Ybus once
        ppci["internal"]["ybus_fact"] = factorized(ppci["internal"]["Ybus"])
        ppci_0["internal"]["ybus_fact"] = factorized(ppci_0["internal"]["Ybus"])

    ppci_bus = _get_is_ppci_bus(net, bus)
    _calc_rx(net, ppci, ppci_bus)
    _add_kappa_to_ppc(net, ppci)

    _calc_rx(net, ppci_0, ppci_bus)
    _calc_ikss_1ph(net, ppci, ppci_0, ppci_bus)

    if net._options["branch_results"]:
        if net._options["fault"] == "3ph":
            _calc_branch_currents_complex(net, ppci, ppci_bus)
        else:
            _calc_branch_currents(net, ppci, ppci_bus)

    ppc_0 = _copy_results_ppci_to_ppc(ppci_0, ppc_0, "sc")
    ppc = _copy_results_ppci_to_ppc(ppci, ppc, "sc")
    _extract_results(net, ppc, ppc_0, bus=bus)
    _clean_up(net)



# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

# Builds the line outage distribution factor matrix.


from numpy import ones, diag, r_, arange
import numpy as np
from scipy.sparse import csr_matrix as sparse

from pandapower.pypower.idx_brch import F_BUS, T_BUS
from pandapower.auxiliary import version_check

try:
    from numba import jit
    version_check('numba')
except ImportError: # pragma: no cover
    from pandapower.pf.no_numba import jit



@jit(nopython=True)
def update_LODF_diag(LODF): # pragma: no cover
    for ix in range(LODF.shape[0]):
        # To preserve the data type of diagnol elments
        LODF[ix, ix] -= (LODF[ix, ix] + 1.)


def makeLODF(branch, PTDF):
    """Builds the line outage distribution factor matrix.

    Returns the DC line outage distribution factor matrix for a given PTDF.
    The matrix is C{nbr x nbr}, where C{nbr} is the number of branches.

    Example::
        H = makePTDF(baseMVA, bus, branch)
        LODF = makeLODF(branch, H)

    @see: L{makePTDF}

    @author: Ray Zimmerman (PSERC Cornell)
    """
    nl, nb = PTDF.shape
    f = np.real(branch[:, F_BUS])
    t = np.real(branch[:, T_BUS])
    Cft = sparse((r_[ones(nl), -ones(nl)],
                      (r_[f, t], r_[arange(nl), arange(nl)])), (nb, nl))

    H = PTDF * Cft
    h = diag(H, 0)
    # Avoid zero division error
    # Implies a N-1 contingency (No backup branch)
    den = (ones((nl, 1)) * h.T * -1 + 1.)

    # Silence warning caused by np.nan
    with np.errstate(divide='ignore', invalid='ignore'):
        LODF = (H / den)

    # LODF = LODF - diag(diag(LODF)) - eye(nl, nl)
    update_LODF_diag(LODF)

    return LODF


def makeOTDF(PTDF, LODF, outage_branches):
    """
    Compute the Outage Transfer Distribution Factors (OTDF) matrix.

    This function creates the OTDF matrix that relates bus power injections
    to branch flows for specified outage scenarios. It's essential that
    outage branches do not lead to isolated nodes or disconnected islands
    in the grid.

    The grid cannot have isolated nodes or disconnected islands. Use the
    pandapower.topology module to identify branches that, if outaged, would
    lead to isolated nodes (determine_stubs) or islands (find_graph_characteristics).

    The resulting matrix has a width equal to the number of nodes and a length
    equal to the number of outage branches multiplied by the total number of branches.
    The dot product of OTDF and the bus power vector in generation reference frame
    (positive for generation, negative for consumption - the opposite of res_bus.p_mw)
    yields an array with outage branch power flows for every outage scenario,
    facilitating the analysis of all outage scenarios under a DC
    power flow approximation.

    Parameters
    ----------
    PTDF : numpy.ndarray
        The Power Transfer Distribution Factor matrix, defining the sensitivity
        of branch flows to bus power injections.
    LODF : numpy.ndarray
        The Line Outage Distribution Factor matrix, describing how branch flows
        are affected by outages of other branches.
    outage_branches : list or numpy.ndarray
        Indices of branches for which outage scenarios are to be considered.

    Returns
    -------
    OTDF : numpy.ndarray
        The Outage Transfer Distribution Factor matrix. Rows correspond to
        outage scenarios, and columns correspond to branch flows.

    Examples
    --------
    >>> H = makePTDF(baseMVA, bus, branch)
    >>> LODF = makeLODF(branch, H)
    >>> outage_branches = [0, 2]  # Example branch indices for outage scenarios
    >>> OTDF = makeOTDF(H, LODF, outage_branches)
    >>> # To obtain a 2D array with the outage results:
    >>> outage_results = (OTDF @ Pbus).reshape(len(outage_branches), -1)

    Notes
    -----
    - The function assumes a DC power flow model.
    - Ensure that the specified outage branches do not lead to grid
      disconnection or isolated nodes.
    """
    OTDF = np.vstack([PTDF + LODF[:, [i]] @ PTDF[[i], :] for i in outage_branches])
    return OTDF


def outage_results_OTDF(OTDF, Pbus, outage_branches):
    """
    Calculate the branch power flows for each outage scenario based on the given
    Outage Transfer Distribution Factors (OTDF), bus power injections (Pbus), and
    specified outage branches.

    This function computes how branch flows are affected under N-1 contingency
    scenarios (i.e., for each branch outage specified). It uses the OTDF matrix and
    the bus power vector (Pbus) to determine the branch flows in each outage scenario.

    Pbus should represent the net power at each bus in the generation reference case.

    Parameters
    ----------
    OTDF : numpy.ndarray
        The Outage Transfer Distribution Factor matrix, which relates bus power
        injections to branch flows under specific outage scenarios. Its shape
        should be (num_outage_scenarios * num_branches, num_buses).
    Pbus : numpy.ndarray
        A vector representing the net power injections at each bus. Positive values
        for generation, negative for consumption. Its length should be equal to
        the total number of buses.
    outage_branches : numpy.ndarray
        An array of indices representing the branches that are outaged in each
        scenario. Its length should be equal to the number of outage scenarios.

    Returns
    -------
    numpy.ndarray
        A 2D array where each row corresponds to an outage scenario and each column
        represents the resulting power flow in a branch. The number of rows is equal
        to the number of outage scenarios, and the number of columns is equal to the
        number of branches.

    Examples
    --------
    >>> OTDF = np.array([...])  # example OTDF matrix
    >>> Pbus = np.array([...])  # example bus power vector
    >>> outage_branches = np.array([...])  # example outage branches
    >>> branch_flows = outage_results_OTDF(OTDF,Pbus,outage_branches)

    Notes
    -----
    The function assumes a linear relationship between bus power injections and
    branch flows, which is typical in DC power flow models.
    """
    # get branch flows as an array first:
    nminus1_otdf = (OTDF @ Pbus.reshape(-1, 1))
    # reshape to a 2D array with rows relating to outage scenarios and columns to
    # the resulting branch power flows
    nminus1_otdf = nminus1_otdf.reshape(outage_branches.shape[0], -1)
    return nminus1_otdf




